using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using UnityEngine.UI;
using UnityEngine.EventSystems;
using static OVRInput;

namespace JayJayWon
{
    public class GameControlUI
    {
        public static void InitUI()
        {
            gameControlDisplayMode = GameControlDisplayModes.standard;
            gridsControlDisplay = new GridsControlDisplay();
            gridsDisplay = new GridsDisplay();

            _nearestGazeUIButton = new NearestUIButton();
            _nearestLeftTouchButton = new NearestUIButton();
            _nearestRightTouchButton = new NearestUIButton();

            _leftHLeapControl = new LeapTouchControl(LeftRight.left);
            _rightHLeapControl = new LeapTouchControl(LeftRight.right);

            RefreshGameControlUIModes();
        }

        public static void AtomNameUpdate(string oldName, string newName)
        {
            GridsDisplay._uiActiveClothingEditor.UpdateAtomNames(oldName, newName);
            GridsDisplay._uiTargetAtomSelector.UpdateAtomNames(oldName, newName);
        }
        public static void AtomRemovedUpdate(Atom atom)
        {
            string atomName = atom.name;
            GridsDisplay._uiActiveClothingEditor.AtomRemovedUpdate(atomName);
            GridsDisplay._uiTargetAtomSelector.AtomRemovedUpdate(atomName);
        }

        public static void OnEnable()
        {
            if (GridsControlDisplay.gridsControlCanvas != null) GridsControlDisplay.gridsControlCanvas._canvas.enabled = true;
            if (gridsDisplay.gridsCanvas != null) gridsDisplay.gridsCanvas._canvas.enabled = true;
        }

        public static void OnDisable()
        {
            if (GridsControlDisplay.gridsControlCanvas != null) GridsControlDisplay.gridsControlCanvas._canvas.enabled = false;
            if (gridsDisplay.gridsCanvas != null) gridsDisplay.gridsCanvas._canvas.enabled = false;
            if (TargetControl._atomVRMoveActive) TargetControl.StopAtomVRMove();
        }

        public static void OnDestroy()
        {
            if (GridsControlDisplay.gridsControlCanvas != null) GridsControlDisplay.gridsControlCanvas.OnDestroy();
            if (gridsDisplay.gridsCanvas != null) gridsDisplay.gridsCanvas.OnDestroy();

            if (TargetControl._atomVRMoveActive) TargetControl.StopAtomVRMove();
        }
        public static void OnSceneLoaded()
        {
            GridsControlDisplay.SetGameControlUIAtPinPosition();
        }

        private static void UpdateHUDText()
        {
            if (RelativePositionComponentUITab.testSpawnModeActive)
            {
                string actionVerb = "Spawn";
                if (RelativePositionComponentUITab.testSpawnMode == TestSpawnMode.teleportAtomMode) actionVerb = "Teleport";
                SuperController.singleton.helpHUDText.text = "UIAssist: Atom " + actionVerb + " Test Mode Active";
            }
            else if (SpawnAtomComponent.atomSpawningActive)
            {
                SuperController.singleton.helpHUDText.text = "UIAssist: Atom Spawning In Progress";
            }
            else if (TargetControl._atomVRMoveActive && !TargetControl._atomVRMoveLeapActivated)
            {
                if (TargetControl._atomVRMovePaused) SuperController.singleton.helpHUDText.text = "UIAssist: Atom Move Mode Paused. Hold Grab button to unpause or click Grab button to deactivate.";
                else SuperController.singleton.helpHUDText.text = "UIAssist: Atom Move Mode Active. Release Grab button to pause.";
            }
            else if (TargetControl._atomVRMoveActive && TargetControl._atomVRMoveLeapActivated)
            {
                if (TargetControl._atomVRMovePaused) SuperController.singleton.helpHUDText.text = "UIAssist: Atom Move Mode Paused. Form OK Hand Gesture to unpause or pinch fingers to deactivate.";
                else SuperController.singleton.helpHUDText.text = "UIAssist: Atom Move Mode Active. Release OK Hand Gesture to pause.";
            }
        }
        public static void RefreshWristUIButtonGrid()
        {
            if (gridsActivated && _currentGrid !=null)
            {
                gridsDisplay.DestroyUIButtons();
                gridsDisplay.CreateUIButtons();
            }
        }
        public static void LaunchWristUIButtonGrid()
        {
            if (_currentGrid == null || _currentGrid._activeButtonCount == 0)
            {
                _currentGrid = null;
                for (int i = UIAStorables.buttonGridsList.Count - 1; i >= 0; i--)
                {
                    if (UIAStorables.buttonGridsList[i]._activeButtonCount > 0) { _currentGrid = UIAStorables.buttonGridsList[i]; }
                }
            }
            if (_currentGrid == null)
            {
                gameControlDisplayMode = GameControlDisplayModes.standard;
                gridsActivated = false;
            }
            else if (UIAGlobals.vrActive && VRSettings.autoPinOnGridOpenJSB.val && !_uiPinned)
            {
                GridsControlDisplay.TogglePinClickFinish();
            }
            RefreshWristUIButtonGrid();
        }

        private static bool isGazeUIControlActive {
            get { return (gazeControlActive || VRSettings.gazeActivateOpenVRGridJSB.val || VRSettings.gazeActivateCloseVRGridJSB.val) && gameControlDisplayMode != GameControlDisplayModes.edit && gameControlUIMode != GameControlUIModes.desktopHUD && gameControlUIMode != GameControlUIModes.vrHUD; }
        }
        private static bool isTouchUIControlActive
        {
            get { return UIAGlobals.vrActive && LeapMotionControlTool.leapMotionControlActiveJSB.val && _leftHLeapControl._indexFingerTipTrans != null && _rightHLeapControl._indexFingerTipTrans != null; }
        }

        private static Ray gazeRay
        {
            get
            {
                Transform cameraT = SuperController.singleton.lookCamera.transform;
                Vector3 eyePosition = cameraT.position - (cameraT.up * 0.1f);
                Ray ray = new Ray(eyePosition, cameraT.forward);
                return ray;
            }
        }

        private static Vector3 leftLeapFingerTip
        {
            get
            {
                return _leftHLeapControl._indexFingerTipTrans.position + (_leftHLeapControl._indexFingerTipTrans.up * UIAConsts._leapFingerTipOffet);
            }
        }

        private static Vector3 rightLeapFingerTip
        {
            get
            {
                return _rightHLeapControl._indexFingerTipTrans.position + (_rightHLeapControl._indexFingerTipTrans.up * UIAConsts._leapFingerTipOffet);
            }
        }

        public static void SetNearestGazeAndLeapButtons(UIDynamicButton uidButton, UIAButtonGrid bg, int buttonRef)
        {
            if (uidButton.gameObject.activeSelf)
            {
                if (isGazeUIControlActive) _nearestGazeUIButton.SetButtonDistance(bg, buttonRef, gazeRay, uidButton);

                if (isTouchUIControlActive)
                {
                    _nearestLeftTouchButton.SetButtonDistance(bg, buttonRef, leftLeapFingerTip, uidButton);
                    _nearestRightTouchButton.SetButtonDistance(bg, buttonRef, rightLeapFingerTip, uidButton);
                }
            }
        }
        public static void SetNearestGazeAndLeapButtons(UIDynamicButton uidButton)
        {
            SetNearestGazeAndLeapButtons(uidButton, null, -1);
        }

        private static void SetNearestGazeAndLeapButtons(UIDynamicToggle uidToggle, UIAButtonGrid bg, int buttonRef)
        {
            if (uidToggle.gameObject.activeSelf)
            {
                if (isGazeUIControlActive) _nearestGazeUIButton.SetToggleDistance(bg, buttonRef, gazeRay, uidToggle);

                if (isTouchUIControlActive)
                {
                    _nearestLeftTouchButton.SetToggleDistance(bg, buttonRef, leftLeapFingerTip, uidToggle);
                    _nearestRightTouchButton.SetToggleDistance(bg, buttonRef, rightLeapFingerTip, uidToggle);
                }
            }
        }
        public static void SetNearestGazeAndLeapButtons(UIDynamicToggle uidToggle)
        {
            SetNearestGazeAndLeapButtons(uidToggle, null, -1);
        }

        public static void RefreshGameControlUIModes()
        {
            if (UIAGlobals.vrActive == false) gameControlUIMode = GameControlUIModes.desktopHUD;
            else if (VRSettings.vrHandControlJSEnum.val == VRHandControl.disabled || (_uiPinned && (VRSettings.vrPinModeJSEnum.val == VRPinMode.hudLMode || VRSettings.vrPinModeJSEnum.val == VRPinMode.hudRMode || VRSettings.vrPinModeJSEnum.val == VRPinMode.hudCustomMode))) gameControlUIMode = GameControlUIModes.vrHUD;
            else if (_uiPinned && VRSettings.vrPinModeJSEnum.val != VRPinMode.disabled) gameControlUIMode = GameControlUIModes.vrStatic;
            else if (VRSettings.vrHandControlJSEnum.val == VRHandControl.vamUI) gameControlUIMode = GameControlUIModes.vrVAMUI;            
            else gameControlUIMode = GameControlUIModes.vrWrist;

            if (gameControlUIMode == GameControlUIModes.vrStatic && VRSettings.gazeMenuSelectModeJSEnum.val == VRGazeUIControlMode.wristAndPinnedUI) gazeControlActive = true;
            else if (gameControlUIMode == GameControlUIModes.vrWrist && VRSettings.gazeMenuSelectModeJSEnum.val != VRGazeUIControlMode.disabled) gazeControlActive = true;
            else gazeControlActive = false;
        }

        private static void DoubleVRSelectCheck(ref float selectTime, int selectMode)
        {
            float currentTime = Time.time;

            if (selectTime == -1f) selectTime = currentTime;
            else
            {
                if (currentTime - selectTime > 0.4f) selectTime = currentTime;
                else
                {
                    selectTime = -1f;
                    if (selectMode == VRControllerDoubleSelectMode.togglePin && gridsControlDisplay.isPinModeAvailable)
                    {
                        GridsControlDisplay._timeLastPinButtonClickDown = currentTime;
                        GridsControlDisplay.TogglePinClickFinish(false);
                    }
                    else if (selectMode != VRControllerDoubleSelectMode.disabled && (selectMode == VRControllerDoubleSelectMode.toggleHide || UIAGlobals.hideGameControlUI))
                    {
                        UIAGlobals.hideGameControlUI = !UIAGlobals.hideGameControlUI;
                    }
                }
            }

        }

        public static void Update()
        {
            // Check if a double click Select is starting or completing - to allow Toggle the pin
            if (SuperController.singleton.GetLeftSelect()) DoubleVRSelectCheck(ref _timeLeftSelect, VRSettings.LcontrollerDoubleSelectModeJSE.val);

            if (SuperController.singleton.GetRightSelect()) DoubleVRSelectCheck(ref _timeRightSelect, VRSettings.RcontrollerDoubleSelectModeJSE.val);

            bool screenshotActive = SuperController.singleton.screenshotCamera.enabled || SuperController.singleton.hiResScreenshotCamera.enabled;

            if (UIAGlobals.hideGameControlUI || screenshotActive)
            {
                GridsControlDisplay.gridsControlCanvas._canvas.gameObject.SetActive(false);
                gridsDisplay.gridsCanvas._canvas.gameObject.SetActive(false);
            }
            else
            {
                GridsControlDisplay.gridsControlCanvas._canvas.gameObject.SetActive(true);
                gridsDisplay.gridsCanvas._canvas.gameObject.SetActive(true);

                if (SuperController.singleton.helpHUDText.text == "") UpdateHUDText();

                if (UIAGlobals.vrActive && LeapMotionControlTool.leapMotionControlActiveJSB.val && _leftHLeapControl._indexFingerTipTrans != null && _rightHLeapControl._indexFingerTipTrans != null)
                {
                    _leftHLeapControl.Update();
                    _rightHLeapControl.Update();
                }
                else
                {
                    _leftHLeapControl.InitialiseFingerTipTransforms();
                    _rightHLeapControl.InitialiseFingerTipTransforms();
                }

                _nearestGazeUIButton.ClearButtonDistance();
                _nearestLeftTouchButton.ClearButtonDistance();
                _nearestRightTouchButton.ClearButtonDistance();
                gridsControlDisplay.Update();
                gridsDisplay.Update();


                // Handle Gaze Button activation
                if (gazeControlActive && gameControlDisplayMode != GameControlDisplayModes.edit)
                {
                    _nearestGazeUIButton.PreActivateButton(0.1f);
                    bool triggerUIButton = false;
                    // Check if Grab activated and trigger gaze action 
                    if (SuperController.singleton.GetLeftGrab() && VRSettings.vrHandControlJSEnum.val == VRHandControl.leftHand && gazeControlActive) triggerUIButton = true;
                    if (SuperController.singleton.GetRightGrab() && VRSettings.vrHandControlJSEnum.val == VRHandControl.rightHand && gazeControlActive) triggerUIButton = true;
                    if (GameControlUI.gameControlUIMode == GameControlUIModes.vrStatic && VRSettings.vrHandControlJSEnum.val == VRHandControl.vamUI && gazeControlActive && (SuperController.singleton.GetRightGrab() || SuperController.singleton.GetLeftGrab())) triggerUIButton = true;

                    if (triggerUIButton && gameControlDisplayMode != GameControlDisplayModes.edit) _nearestGazeUIButton.ActivateButton(0.1f);
                }

                // Handle Leap Touch Button activation
                if (UIAGlobals.vrActive && LeapMotionControlTool.leapMotionControlActiveJSB.val)
                {
                    _nearestLeftTouchButton.PreActivateButton(0f, 60f, 30f);
                    _nearestRightTouchButton.PreActivateButton(0f, 60f, 30f);
                    _nearestLeftTouchButton.ActivateButton(0f, 5f, true, LeftRight.left);
                    _nearestRightTouchButton.ActivateButton(0f, 5f, true, LeftRight.right);
                }

                // Handle Gaze Open or Close Grid
                if ((VRSettings.gazeActivateOpenVRGridJSB.val || VRSettings.gazeActivateCloseVRGridJSB.val) && gameControlUIMode != GameControlUIModes.vrVAMUI && gameControlUIMode != GameControlUIModes.desktopHUD && gameControlUIMode != GameControlUIModes.vrHUD && gameControlDisplayMode != GameControlDisplayModes.edit)
                {
                    gridsControlDisplay.GazeOpenOrCloseGrid();
                }
            }

            // Check and handle if VR Move Atom is active 
            if (TargetControl._atomVRMoveActive) TargetControl.UpdateAtomMove();
        }

        public static int gameControlUIMode { get; private set; }
        public static bool gazeControlActive { get; private set; }

        public static NearestUIButton _nearestGazeUIButton;
        public static NearestUIButton _nearestLeftTouchButton;
        public static NearestUIButton _nearestRightTouchButton;

        public static LeapTouchControl _leftHLeapControl;
        public static LeapTouchControl _rightHLeapControl;

        public static GridsDisplay gridsDisplay;
        public static GridsControlDisplay gridsControlDisplay;

        public static bool _uiPinned = false;
        public static bool gridsActivated = false;

        public static bool _isUIGazeDeactivated = true;
        public static UIAButtonGrid _currentGrid ;
        private static float _timeLeftSelect = -1f;
        private static float _timeRightSelect = -1f;
        public static int gameControlDisplayMode { get; set; }
    }
    public class GridsControlDisplay
    {
        public static CanvasBase gridsControlCanvas;
        public static VAMUILaunchCanvas launchButtonCanvas;

        public static UIDynamicButton _mainButton;
        public static UIDynamicButton vamUILaunchButton;

        public static UIDynamicButton _leftButton;
        public static UIDynamicButton _rightButton;
        public static UIDynamicButton _firstButton;
        public static UIDynamicButton _lastButton;
        public static UIDynamicButton _mainPinButton;
        public static float _timeLastMainButtonClickDown = 0f;
        public static float _timeLastPinButtonClickDown = 0f;
        private static bool _pinButtonMoveModeActive = false;
        private static bool _pinIgnoreNextClickUp = false;

        private static float timeGazeOpenStarted = -1f;

        public static int _vrPinMoveControllerLorR = LeftRight.neither;
        public static Vector3 _vrPinMoveButtonOffsetPosition = new Vector3();
        private static bool _autoDePinEnabled = false;
        public static Vector3 _fixedPinPosition;

        public GridsControlDisplay()
        {
            gridsControlCanvas = new CanvasBase();
            launchButtonCanvas = new VAMUILaunchCanvas();

            _mainButton = gridsControlCanvas.CreateUIButtonInCanvas(System.DateTime.Now.ToString("HH:mm"), 0f, 0f, 0f, 70f, 40f);
            _mainButton.button.onClick.AddListener(() => { ToggleUIActivateButton(); });
            EventTrigger trigger = _mainButton.button.gameObject.AddComponent<EventTrigger>();
            var pointerDown = new EventTrigger.Entry();
            pointerDown.eventID = EventTriggerType.PointerDown;
            pointerDown.callback.AddListener((e) => MainButtonClickDown());
            trigger.triggers.Add(pointerDown);

            var pointerUp = new EventTrigger.Entry();
            pointerUp.eventID = EventTriggerType.PointerUp;
            pointerUp.callback.AddListener((e) => MainButtonClickUp());
            trigger.triggers.Add(pointerUp);

            vamUILaunchButton = launchButtonCanvas.CreateUIButtonInCanvas("UI Assist",380f,-65f , 0f, 70f , 40f );
            vamUILaunchButton.button.onClick.AddListener(() => {
                ToggleUIActivateButton();
                SuperController.singleton.activeUI = SuperController.ActiveUI.None;
                if (BAInterop.isBAenabled && !GameControlUI.gridsActivated)
                {
                    BAInterop.BAPlugin.SendMessage("OnCloseVAMUIAttachedUI",SendMessageOptions.DontRequireReceiver);
                }
            });
            ResetVAMUILaunchButtonPosition();

            EventTrigger trigger2 = vamUILaunchButton.button.gameObject.AddComponent<EventTrigger>();
            var pointerDown2 = new EventTrigger.Entry();
            pointerDown2.eventID = EventTriggerType.PointerDown;
            pointerDown2.callback.AddListener((e) => MainButtonClickDown());
            trigger2.triggers.Add(pointerDown2);

            var pointerUp2 = new EventTrigger.Entry();
            pointerUp2.eventID = EventTriggerType.PointerUp;
            pointerUp2.callback.AddListener((e) => MainButtonClickUp());
            trigger2.triggers.Add(pointerUp);


            _leftButton = gridsControlCanvas.CreateUIButtonInCanvas("<<", -50f, 0f, 0f, 30f, 30f);
            _leftButton.gameObject.SetActive(false);
            _leftButton.button.onClick.AddListener(() => { ScrollLeftButton(); });
            _rightButton = gridsControlCanvas.CreateUIButtonInCanvas(">>", 50f, 0f, 0f, 30f, 30f);
            _rightButton.gameObject.SetActive(false);
            _rightButton.button.onClick.AddListener(() => { ScrollRightButton(); });

            _firstButton = gridsControlCanvas.CreateUIButtonInCanvas("|<", -80f, 0f, 0f, 30f, 30f);
            _firstButton.gameObject.SetActive(false);
            _firstButton.button.onClick.AddListener(() => { ScrollFirstButton(); });
            _lastButton = gridsControlCanvas.CreateUIButtonInCanvas(">|", 80f, 0f, 0f, 30f, 30f);
            _lastButton.gameObject.SetActive(false);
            _lastButton.button.onClick.AddListener(() => { ScrollLastButton(); });

            _mainPinButton = gridsControlCanvas.CreateUIButtonInCanvas("O", -120f, 0f, 0f, 30f, 30f);
            _mainPinButton.gameObject.SetActive(false);

            _leftButton.buttonColor = Color.grey;
            _rightButton.buttonColor = Color.grey;
            _firstButton.buttonColor = Color.grey;
            _lastButton.buttonColor = Color.grey;
            _mainPinButton.buttonColor = Color.grey;

            EventTrigger triggerTogglePin = _mainPinButton.button.gameObject.AddComponent<EventTrigger>();
            var pointerDownPin = new EventTrigger.Entry();
            pointerDownPin.eventID = EventTriggerType.PointerDown;
            pointerDownPin.callback.AddListener((e) => TogglePinClickDown());
            triggerTogglePin.triggers.Add(pointerDownPin);

            var pointerUpPin = new EventTrigger.Entry();
            pointerUpPin.eventID = EventTriggerType.PointerUp;
            pointerUpPin.callback.AddListener((e) => TogglePinClickUp());
            triggerTogglePin.triggers.Add(pointerUpPin);

            if (gridsControlCanvas.attachedVAMUI) ResetButtonPositions();
        }

        public static void OnDestroy()
        {
            if (SuperController.singleton != null)
            {
                SuperController.singleton.RemoveCanvas(launchButtonCanvas._canvas);
            }
            if (launchButtonCanvas != null) launchButtonCanvas._canvas.transform.SetParent(null, false);
            if (launchButtonCanvas._canvas.gameObject != null) GameObject.Destroy(launchButtonCanvas._canvas.gameObject);
        }

        public static void ResetVAMUILaunchButtonPosition()
        {
            if (UIAGlobals.vrActive && vamUILaunchButton!=null)
            {
                float xPos = 380f;
                float yPos = -65f;
                if (VRSettings.vrUIAttachPositionJSEnum.val== VAMUIattachPosition.leftBottom) xPos = -380f ;

                vamUILaunchButton.button.transform.localPosition = new Vector3(xPos, yPos, 0.0f);
            }
            
        }

        public static void ResetButtonPositions()
        {

            float widthOffset = 0f;
            float heightOffset = 0f;
            if (gridsControlCanvas.attachedVAMUI)
            {
                widthOffset = 170f;
                heightOffset =0f;
            }

            _leftButton.button.transform.localPosition = new Vector3(-widthOffset - 50f, heightOffset, 0f);
            _rightButton.button.transform.localPosition = new Vector3(widthOffset + 50f, heightOffset, 0f);

            _firstButton.button.transform.localPosition = new Vector3(-widthOffset - 80f, heightOffset, 0f);
            _lastButton.button.transform.localPosition = new Vector3(widthOffset + 80f, heightOffset, 0f);

            _mainPinButton.button.transform.localPosition = new Vector3(-widthOffset - 120f, heightOffset, 0f);

        }

        private static void UpdateLeftRightButtons()
        {
            if (!GameControlUI.gridsActivated || GameControlUI.gameControlDisplayMode == GameControlDisplayModes.ace || GameControlUI.gameControlDisplayMode == GameControlDisplayModes.clothItemPresetSelect || GameControlUI.gameControlDisplayMode == GameControlDisplayModes.atomSelect || (GameControlUI.gameControlDisplayMode == GameControlDisplayModes.edit && (GridsDisplay._uiMainButtonGridDisplay.IsButtonMoveActive() || GridsDisplay._uiQLBButtonGridDisplay.IsButtonMoveActive())))
            {
                _leftButton.gameObject.SetActive(false);
                _rightButton.gameObject.SetActive(false);
                _firstButton.gameObject.SetActive(false);
                _lastButton.gameObject.SetActive(false);
                if (GameControlUI.gameControlUIMode == GameControlUIModes.vrVAMUI && !GameControlUI.gridsActivated) _mainPinButton.gameObject.SetActive(false);
            }
            else
            {
                if (GetNextLeftGrid() >= 0)
                {
                    _leftButton.gameObject.SetActive(true);
                    _firstButton.gameObject.SetActive(true);
                }
                else
                {
                    _leftButton.gameObject.SetActive(false);
                    _firstButton.gameObject.SetActive(false);
                }
                if (GetNextRightGrid() >= 0)
                {
                    _rightButton.gameObject.SetActive(true);
                    _lastButton.gameObject.SetActive(true);
                }
                else
                {
                    _rightButton.gameObject.SetActive(false);
                    _lastButton.gameObject.SetActive(false);
                }
                if (GameControlUI.gameControlUIMode!=GameControlUIModes.desktopHUD) _mainPinButton.gameObject.SetActive(true);
                else _mainPinButton.gameObject.SetActive(false);
            }
        }

        public static void TogglePinClickDown(int leapVRHand = LeftRight.neither)
        {
            try
            {
                _timeLastPinButtonClickDown = Time.time;

                if (!GameControlUI._uiPinned)
                {
                    TogglePinClickFinish();
                    _pinIgnoreNextClickUp = true;
                }
                else if (!_pinButtonMoveModeActive)
                {
                    _pinButtonMoveModeActive = true;
                    Vector3[] rectCornersArray = new Vector3[4];

                    RectTransform rt = _mainPinButton.transform.GetComponent<RectTransform>();
                    rt.GetWorldCorners(rectCornersArray);
                    Vector3 pinButtonCentreWC = rectCornersArray[0] + ((rectCornersArray[2] - rectCornersArray[0]) * 0.5f);

                    if (leapVRHand == LeftRight.neither)
                    {
                        int vrHand = LeftRight.neither;
                        if (SuperController.singleton.GetRightSelect()) vrHand = LeftRight.right;
                        if (SuperController.singleton.GetLeftSelect()) vrHand = LeftRight.left;

                        Transform vrHandRayTransform = null;

                        if (vrHand == LeftRight.left) vrHandRayTransform = SuperController.singleton.leftControllerCamera.transform.transform;
                        else vrHandRayTransform = SuperController.singleton.rightControllerCamera.transform.transform;

                        _vrPinMoveControllerLorR = vrHand;
                        _vrPinMoveButtonOffsetPosition = vrHandRayTransform.InverseTransformPoint(pinButtonCentreWC);
                    }
                    else
                    {
                        Transform vrLeapHandTransform = null;
                        if (leapVRHand == LeftRight.left) vrLeapHandTransform = GameControlUI._leftHLeapControl._thumbTipTrans;
                        else vrLeapHandTransform = GameControlUI._rightHLeapControl._thumbTipTrans;

                        _vrPinMoveControllerLorR = leapVRHand;
                        _vrPinMoveButtonOffsetPosition = vrLeapHandTransform.InverseTransformPoint(pinButtonCentreWC);
                    }
                }
                else
                {
                    _pinButtonMoveModeActive = false;
                    _pinIgnoreNextClickUp = true;
                }

            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public static void TogglePinClickFinish(bool closeUIonDePin = true)
        {
            

            if (Time.time - _timeLastPinButtonClickDown < 0.6f || !GameControlUI._uiPinned)
            {
                if (GameControlUI._uiPinned)
                {
                    if (GameControlUI.gameControlDisplayMode != GameControlDisplayModes.edit && closeUIonDePin)
                    {
                        CloseGrids();
                        GameControlUI._isUIGazeDeactivated = false;
                    }

                    _pinButtonMoveModeActive = false;
                }
                GameControlUI._uiPinned = !GameControlUI._uiPinned;

                if (GameControlUI._uiPinned)
                {
                    SetGameControlUIAtPinPosition();
                    _autoDePinEnabled = false;
                }
            }
            else if (GameControlUI._uiPinned)
            {
                _pinButtonMoveModeActive = false;
            }
            GameControlUI.RefreshGameControlUIModes();
        }

        public static void SetGameControlUIAtPinPosition()
        {
            float worldScale = SuperController.singleton.worldScale;
            if (!UIAGlobals.vrActive) worldScale = 1f;

            if (VRSettings.vrHandControlJSEnum.val == VRHandControl.rightHand) _fixedPinPosition = SuperController.singleton.rightHand.position;
            else _fixedPinPosition = SuperController.singleton.leftHand.position;
            Vector2 fixedPinPos2D = new Vector2(_fixedPinPosition.x, _fixedPinPosition.z);
            Vector2 vrHeadPos2D = new Vector2(CameraTarget.centerTarget.transform.position.x, CameraTarget.centerTarget.transform.position.z);
            if (Vector2.Distance(fixedPinPos2D, vrHeadPos2D) < VRSettings.vrPinMinOffsetJSF.val * worldScale)
            {
                Vector2 pinOffset = fixedPinPos2D - vrHeadPos2D;
                Vector2 targetPosition = vrHeadPos2D + (pinOffset.normalized * VRSettings.vrPinMinOffsetJSF.val * SuperController.singleton.worldScale);
                _fixedPinPosition.x = targetPosition.x;
                _fixedPinPosition.z = targetPosition.y;
            }
        }

        private static void TogglePinClickUp()
        {
            try
            {
                if (UIAGlobals.vrActive && !_pinIgnoreNextClickUp)
                {
                    TogglePinClickFinish();
                }
                else _pinIgnoreNextClickUp = false;
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }

        public static void MainButtonClickDown()
        {
            try
            {
                _timeLastMainButtonClickDown = Time.time;
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }

        public static void MainButtonClickFinish()
        {
            if ((GameControlUI.gameControlDisplayMode == GameControlDisplayModes.clothItemPresetSelect || GameControlUI.gameControlDisplayMode == GameControlDisplayModes.atomSelect || GameControlUI.gameControlDisplayMode == GameControlDisplayModes.ace) && GameControlUI.gameControlUIMode == GameControlUIModes.vrVAMUI && !GameControlUI.gridsActivated)
            {
                GameControlUI.gridsActivated = true;
                RefreshGridButtons();
            }
            else if (GameControlUI.gameControlDisplayMode == GameControlDisplayModes.ace)
            {
                GameControlUI.gameControlDisplayMode = GameControlDisplayModes.standard;
                GameControlUI.RefreshWristUIButtonGrid();

            }
            else if (GameControlUI.gameControlDisplayMode == GameControlDisplayModes.clothItemPresetSelect)
            {
                GameControlUI.gameControlDisplayMode = GameControlDisplayModes.ace;
                GameControlUI.RefreshWristUIButtonGrid();

            }
            else if (GameControlUI.gameControlDisplayMode == GameControlDisplayModes.atomSelect)
            {
                GameControlUI.gameControlDisplayMode = GameControlDisplayModes.standard;
                GridsDisplay._uiTargetAtomSelector.DestroyUI();
                GameControlUI.RefreshWristUIButtonGrid();
            }
            else if (!_pinButtonMoveModeActive)
            {
                if (Time.time - _timeLastMainButtonClickDown > 0.4f)
                {
                    GameControlUI.gameControlDisplayMode = GameControlDisplayModes.edit;
                    GameControlUI.gridsActivated = true;
                }
                else if (GameControlUI.gameControlDisplayMode == GameControlDisplayModes.edit) GameControlUI.gameControlDisplayMode = GameControlDisplayModes.standard;
                else GameControlUI.gridsActivated = !GameControlUI.gridsActivated;

                RefreshGridButtons();

            }

        }

        public static void RefreshGridButtons()
        {
            if (!GameControlUI.gridsActivated)
            {
                GameControlUI.gridsDisplay.DestroyUIButtons();
                GameControlUI._isUIGazeDeactivated = false;
                if (VRSettings.autoUnPinOnGridCloseJSB.val)
                {
                    GameControlUI._uiPinned = false;
                    GameControlUI.RefreshGameControlUIModes();
                }
            }
            else GameControlUI.LaunchWristUIButtonGrid();
        }

        private static void MainButtonClickUp()
        {
            try
            {
                if (UIAGlobals.vrActive)
                {
                    MainButtonClickFinish();
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }

        private static void ToggleUIActivateButton()
        {
            try
            {
                if (!UIAGlobals.vrActive)
                {
                    MainButtonClickFinish();
                }

            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public static int GetNextLeftGridWithBlankButton()
        {
            int nextGrid = -1;
            for (int i = GameControlUI._currentGrid.gridIndex - 1; i >= 0; i--)
            {
                if (UIAStorables.buttonGridsList[i].ContainsBlankButtons())
                {
                    nextGrid = i;
                    break;
                }
            }
            return (nextGrid);
        }
        public static int GetNextRightGridWithBlankButton()
        {
            int nextGrid = -1;
            for (int i = GameControlUI._currentGrid.gridIndex + 1; i < UIAStorables.buttonGridsList.Count; i++)
            {
                if (UIAStorables.buttonGridsList[i].ContainsBlankButtons())
                {
                    nextGrid = i;
                    break;
                }
            }
            return (nextGrid);
        }
        public static int GetMaxRightGridWithActiveButtons()
        {
            int nextGrid = -1;
            for (int i = UIAStorables.buttonGridsList.Count - 1; i > -1; i--)
            {
                if (UIAStorables.buttonGridsList[i]._activeButtonCount > 0)
                {
                    nextGrid = i;
                    break;
                }
            }
            return (nextGrid);
        }
        public static int GetMinLeftGridWithActiveButtons()
        {
            int nextGrid = -1;
            for (int i = 0; i < UIAStorables.buttonGridsList.Count; i++)
            {
                if (UIAStorables.buttonGridsList[i]._activeButtonCount > 0)
                {
                    nextGrid = i;
                    break;
                }
            }
            return (nextGrid);
        }
        private static int GetNextLeftGrid()
        {
            int nextGrid = -1;
            for (int i = 0; i < GameControlUI._currentGrid.gridIndex; i++)
            {
                if (UIAStorables.buttonGridsList[i]._activeButtonCount > 0) { nextGrid = i; }
            }
            return (nextGrid);
        }
        private static int GetNextRightGrid()
        {
            int nextGrid = -1;
            for (int i = UIAStorables.buttonGridsList.Count - 1; i > GameControlUI._currentGrid.gridIndex; i--)
            {
                if (UIAStorables.buttonGridsList[i]._activeButtonCount > 0) { nextGrid = i; }
            }
            return (nextGrid);
        }
        public static void SetCurrentGrid(int gridIndex)
        {
            try
            {
                if (gridIndex == -1) GameControlUI._currentGrid = null;
                else GameControlUI._currentGrid = UIAStorables.buttonGridsList[gridIndex];
                GameControlUI.gridsDisplay.DestroyUIButtons();
                if (GameControlUI._currentGrid != null)
                {
                    GameControlUI.gridsDisplay.CreateUIButtons();
                    GameControlUI.gridsActivated = true;
                }
                else CloseGrids();
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public static void ScrollFirstButton()
        {
            try
            {
                SetCurrentGrid(GetMinLeftGridWithActiveButtons());
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public static void ScrollLastButton()
        {
            try
            {
                SetCurrentGrid(GetMaxRightGridWithActiveButtons());
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public static void ScrollLeftButton()
        {
            try
            {
                GameControlUI._currentGrid = UIAStorables.buttonGridsList[GetNextLeftGrid()];
                GameControlUI.gridsDisplay.DestroyUIButtons();
                if (GameControlUI._currentGrid!=null)
                {
                    GameControlUI.gridsDisplay.CreateUIButtons();
                    GameControlUI.gridsActivated = true;
                }
                else { CloseGrids(); }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public static void ScrollRightButton()
        {
            try
            {
                GameControlUI._currentGrid = UIAStorables.buttonGridsList[GetNextRightGrid()] ;
                GameControlUI.gridsDisplay.DestroyUIButtons();
                if (GameControlUI._currentGrid !=null)
                {
                    GameControlUI.gridsDisplay.CreateUIButtons();
                    GameControlUI.gridsActivated = true;
                }
                else { CloseGrids(); }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }

        public bool isPinModeAvailable
        {
            get
            {
                if (GameControlUI.gameControlUIMode == GameControlUIModes.desktopHUD || VRSettings.vrPinModeJSEnum.val == VRPinMode.disabled || VRSettings.vrHandControlJSEnum.val == VRHandControl.disabled) return false;
                if (!GameControlUI.gridsActivated && GameControlUI.gameControlUIMode == GameControlUIModes.vrWrist) return false;
                if (GameControlUI.gameControlDisplayMode == GameControlDisplayModes.edit && (GridsDisplay._uiMainButtonGridDisplay.IsButtonMoveActive() || GridsDisplay._uiQLBButtonGridDisplay.IsButtonMoveActive())) return false;
                if (UIAGlobals.hideGameControlUI) return false;

                return true;
            }
        }


        private void UpdateButtonDisplay()
        {
            if (_mainPinButton.gameObject.activeSelf && GameControlUI._uiPinned) _mainPinButton.label = "O";
            else if (_mainPinButton.gameObject.activeSelf && !GameControlUI._uiPinned) _mainPinButton.label = "O";

            if (GameControlUI.gameControlDisplayMode == GameControlDisplayModes.edit)
            {
                _mainButton.label = "EDIT MODE";
                _mainButton.buttonText.fontSize = 14;
                _mainButton.textColor = Color.white;
                _mainButton.buttonColor = Color.red;

                vamUILaunchButton.label = "EDIT MODE";
                vamUILaunchButton.buttonText.fontSize = 14;
                vamUILaunchButton.textColor = Color.white;
                vamUILaunchButton.buttonColor = Color.red;
            }
            else if (GameControlUI.gameControlDisplayMode == GameControlDisplayModes.clothItemPresetSelect)
            {
                _mainButton.label = "CLOTHING ITEM PRESETS";
                _mainButton.buttonText.fontSize = 11;
                _mainButton.textColor = Color.black;
                _mainButton.buttonColor = Color.green;

                vamUILaunchButton.label = "CLOTHING ITEM PRESETS";
                vamUILaunchButton.buttonText.fontSize = 11;
                vamUILaunchButton.textColor = Color.black;
                vamUILaunchButton.buttonColor = Color.green;
            }
            else if (GameControlUI.gameControlDisplayMode == GameControlDisplayModes.ace)
            {
                _mainButton.label = "ACTIVE CLOTHING";
                _mainButton.buttonText.fontSize = 12;
                _mainButton.textColor = Color.black;
                _mainButton.buttonColor = Color.green;

                vamUILaunchButton.label = "ACTIVE CLOTHING";
                vamUILaunchButton.buttonText.fontSize = 12;
                vamUILaunchButton.textColor = Color.black;
                vamUILaunchButton.buttonColor = Color.green;
            }
            else if (GameControlUI.gameControlDisplayMode == GameControlDisplayModes.atomSelect)
            {
                _mainButton.label = "SELECT TARGET ATOM";
                _mainButton.buttonText.fontSize = 11;
                _mainButton.textColor = Color.black;
                _mainButton.buttonColor = Color.cyan;

                vamUILaunchButton.label = "SELECT TARGET ATOM";
                vamUILaunchButton.buttonText.fontSize = 11;
                vamUILaunchButton.textColor = Color.black;
                vamUILaunchButton.buttonColor = Color.cyan;
            }
            else if (GameControlUI.gridsActivated && GameControlUI._currentGrid.gridLabel != "" && GameControlSettings.displayGridLabelsJSB.val)
            {
                _mainButton.label = GameControlUI._currentGrid.gridLabel;
                _mainButton.buttonText.fontSize = 16;
                _mainButton.textColor = Color.black;
                _mainButton.buttonColor = Color.grey;

                vamUILaunchButton.label = GameControlUI._currentGrid.gridLabel;
                vamUILaunchButton.buttonText.fontSize = 16;
                vamUILaunchButton.textColor = Color.black;
                vamUILaunchButton.buttonColor = Color.grey;
            }
            else
            {
                if (!GameControlSettings.i24hClockEnableJSB.val)
                {
                    _mainButton.label = System.DateTime.Now.ToString("h:mm tt");
                    vamUILaunchButton.label = System.DateTime.Now.ToString("h:mm tt");
                }
                else
                {
                    _mainButton.label = System.DateTime.Now.ToString("HH:mm");
                    vamUILaunchButton.label = System.DateTime.Now.ToString("HH:mm");
                }
                _mainButton.buttonText.fontSize = 16;
                _mainButton.textColor = Color.black;
                _mainButton.buttonColor = Color.grey;
                vamUILaunchButton.buttonText.fontSize = 16;
                vamUILaunchButton.textColor = Color.black;
                vamUILaunchButton.buttonColor = Color.grey;
            }
        }

        public void GazeOpenOrCloseGrid()
        {
            float currentTime = Time.time;
            Transform cameraT = SuperController.singleton.lookCamera.transform;
            Vector3 mainButtonToEyeDirection = _mainPinButton.transform.position - cameraT.position;
            Vector3 eyeToMainButtonDirection = cameraT.position -_mainPinButton.transform.position;
            float mainButtonToEyeAngle = Vector3.Angle(_mainPinButton.transform.forward, mainButtonToEyeDirection);
            float eyeToMainButtonAngle = Vector3.Angle(eyeToMainButtonDirection, -cameraT.forward);

            if (!GameControlUI.gridsActivated && GameControlUI._nearestGazeUIButton.GetDistance() < 0.2f && mainButtonToEyeAngle < 25f && eyeToMainButtonAngle < 30f && GameControlUI._isUIGazeDeactivated && VRSettings.gazeActivateOpenVRGridJSB.val)
            {
                if (timeGazeOpenStarted == -1f) timeGazeOpenStarted = currentTime;
                else if (currentTime - timeGazeOpenStarted >0.2f)
                {
                    GameControlUI.gridsActivated = true;
                    GameControlUI.LaunchWristUIButtonGrid();
                    timeGazeOpenStarted = -1f;
                }
                
            }
            else timeGazeOpenStarted = -1f;


            if (GameControlUI.gridsActivated && (GameControlUI._nearestGazeUIButton.GetDistance() > 0.25f || mainButtonToEyeAngle > 90f) && !_pinButtonMoveModeActive && VRSettings.gazeActivateCloseVRGridJSB.val)
            {
                CloseGrids();
                if (VRSettings.autoUnPinOnGridCloseJSB.val)
                {
                    GameControlUI._uiPinned = false;
                    GameControlUI.RefreshGameControlUIModes();
                }
                GameControlUI._isUIGazeDeactivated = true;

            }
            if (!GameControlUI.gridsActivated && ((GameControlUI._nearestGazeUIButton.GetDistance() > 0.25f) || mainButtonToEyeAngle > 90f)) GameControlUI._isUIGazeDeactivated = true;
        }

        public static void CloseGrids()
        {
            GameControlUI.gridsActivated = false;
            GridsControlDisplay.RefreshGridButtons();
            
        }
        private void UpdateButtonDistances()
        {
            if (GameControlUI.gameControlUIMode == GameControlUIModes.vrVAMUI) GameControlUI.SetNearestGazeAndLeapButtons(vamUILaunchButton);
            else GameControlUI.SetNearestGazeAndLeapButtons(_mainButton);
            GameControlUI.SetNearestGazeAndLeapButtons(_leftButton);
            GameControlUI.SetNearestGazeAndLeapButtons(_rightButton);
            GameControlUI.SetNearestGazeAndLeapButtons(_firstButton);
            GameControlUI.SetNearestGazeAndLeapButtons(_lastButton);
            GameControlUI.SetNearestGazeAndLeapButtons(_mainPinButton);
        }

        public void Update()
        {
            gridsControlCanvas.Update();
            launchButtonCanvas.Update();

            if (GameControlUI.gameControlUIMode == GameControlUIModes.vrVAMUI)
            {
                _mainButton.gameObject.SetActive(false);
                vamUILaunchButton.gameObject.SetActive(true);
            }
            else
            {
                _mainButton.gameObject.SetActive(true);
                vamUILaunchButton.gameObject.SetActive(false);
            }

            UpdateButtonDistances();
            UpdateButtonDisplay();

            if (GameControlUI.gameControlUIMode == GameControlUIModes.vrStatic && VRSettings.vrHandControlJSEnum.val != VRHandControl.disabled && VRSettings.vrHandControlJSEnum.val != VRHandControl.vamUI)
            {
                Transform vrHandTransform = SuperController.singleton.leftHand;
                if (VRSettings.vrHandControlJSEnum.val == VRHandControl.rightHand) vrHandTransform = SuperController.singleton.rightHand;

                float distance = Vector3.Distance(_fixedPinPosition, vrHandTransform.position);
                if (distance > 0.2f) _autoDePinEnabled = true;
                if (distance < 0.1f & _autoDePinEnabled)
                {
                    _timeLastPinButtonClickDown = Time.time;
                    TogglePinClickFinish(false);
                }
            }

            // Check if pin Mode is available
            if (isPinModeAvailable) _mainPinButton.gameObject.SetActive(true);
            else _mainPinButton.gameObject.SetActive(false);

            if (!GameControlUI.gridsActivated && UIAGlobals.vrActive && GameControlUI.gameControlUIMode == GameControlUIModes.vrWrist) gridsControlCanvas._canvasGroup.alpha = VRSettings.vrWristOpacityJSF.val;
            else gridsControlCanvas._canvasGroup.alpha = 1f;

            UpdateLeftRightButtons();

            if (_pinButtonMoveModeActive && GameControlUI._uiPinned)
            {
                Transform vrMoveHandTransform = null;
                if (_vrPinMoveControllerLorR == LeftRight.left)
                {
                    if (GameControlUI._leftHLeapControl._pinMoveModeActive) vrMoveHandTransform = GameControlUI._leftHLeapControl._thumbTipTrans;
                    else vrMoveHandTransform = SuperController.singleton.leftControllerCamera.transform.transform;
                }
                else if (_vrPinMoveControllerLorR == LeftRight.right)
                {
                    if (GameControlUI._rightHLeapControl._pinMoveModeActive) vrMoveHandTransform = GameControlUI._rightHLeapControl._thumbTipTrans;
                    else vrMoveHandTransform = SuperController.singleton.rightControllerCamera.transform.transform;
                }

                Vector3[] rectCornersArray = new Vector3[4];
                RectTransform rt = _mainPinButton.transform.GetComponent<RectTransform>();
                rt.GetWorldCorners(rectCornersArray);
                Vector3 buttonCentre = rectCornersArray[0] + ((rectCornersArray[2] - rectCornersArray[0]) * 0.5f);
                Vector3 buttonCanvasOffset = _fixedPinPosition - buttonCentre;

                Vector3 targetButtonPostion = vrMoveHandTransform.TransformPoint(_vrPinMoveButtonOffsetPosition);
                _fixedPinPosition = targetButtonPostion + buttonCanvasOffset;
            }
        }
        
    }

    public class GridsDisplay
    {
        public GridsCanvas gridsCanvas;

        public GameObject _gasButtonScrollListGO;
        private GameObject _gasHideControlNodesToggleGO;
        private List<GameObject> _gasScrollListButtonListGOs;
        public List<GazeTarget> _orderedGazeAssistSelectionTargets = new List<GazeTarget>();

        public static SingleGridDisplay _uiMainButtonGridDisplay;
        public static SingleGridDisplay _uiQLBButtonGridDisplay;
        public static ACEPlusDisplay _uiActiveClothingEditor;
        public static TargetAtomSelectorDisplay _uiTargetAtomSelector;
        public static ClothItemPresetSelector _uiClothItemPresetSelector;
        public static UIDynamicTextField _gazeStatusTextField;

        public UIDynamicTextField _moveUIDButtonToLeftGrid;
        public UIDynamicTextField _moveUIDButtonToRightGrid;

        public bool _moveUIDButtonToLeftGridInRange = false;
        public bool _moveUIDButtonToRightGridInRange = false;

        public GridsDisplay()
        {
            gridsCanvas = new GridsCanvas();
            _uiMainButtonGridDisplay = new SingleGridDisplay(gridsCanvas);
            _uiQLBButtonGridDisplay = new SingleGridDisplay(gridsCanvas);
            _uiActiveClothingEditor = new ACEPlusDisplay(gridsCanvas);
            _uiClothItemPresetSelector = new ClothItemPresetSelector(gridsCanvas);

            _gasScrollListButtonListGOs = new List<GameObject>();
            _uiTargetAtomSelector = new TargetAtomSelectorDisplay(gridsCanvas);
        }

        private void UpdateButtonDistances()
        {
            if (GameControlUI.gameControlDisplayMode == GameControlDisplayModes.standard || GameControlUI.gameControlDisplayMode == GameControlDisplayModes.edit)
            {
                _uiMainButtonGridDisplay.UpdateButtonDistances();

                if (GameControlSettings.activateQuickLaunchBarJSB.val) _uiQLBButtonGridDisplay.UpdateButtonDistances();

                if (_gasHideControlNodesToggleGO != null)
                {
                    UIDynamicToggle uiToggle = _gasHideControlNodesToggleGO.transform.GetComponent<UIDynamicToggle>();
                    GameControlUI.SetNearestGazeAndLeapButtons(uiToggle);
                }

                if (_gasButtonScrollListGO != null)
                {
                    RectTransform scrollListRT = _gasButtonScrollListGO.GetComponent<RectTransform>();
                    float checkRectMinY = scrollListRT.rect.yMin;
                    float checkRectMaxY = scrollListRT.rect.yMax;

                    for (int index = 0; index < _gasScrollListButtonListGOs.Count; index++)
                    {

                        UIDynamicButton selectTargetButton = _gasScrollListButtonListGOs[index].transform.GetComponent<UIDynamicButton>();

                        RectTransform scrollListButtonRT = _gasScrollListButtonListGOs[index].GetComponent<RectTransform>();
                        Vector3 positionInWorld = scrollListButtonRT.parent.TransformPoint(scrollListButtonRT.localPosition);
                        Vector3 positionInScroll = scrollListRT.InverseTransformPoint(positionInWorld);

                        float childMinY = positionInScroll.y + scrollListButtonRT.rect.yMin;
                        float childMaxY = positionInScroll.y + scrollListButtonRT.rect.yMax;

                        if (childMaxY >= checkRectMinY && childMinY <= checkRectMaxY)
                        {
                            GameControlUI.SetNearestGazeAndLeapButtons(selectTargetButton);
                        }


                    }

                }

            }
            else if (GameControlUI.gameControlDisplayMode == GameControlDisplayModes.ace)
            {
                _uiActiveClothingEditor.UpdateACEGazeTouchDistances();
            }
            else if (GameControlUI.gameControlDisplayMode == GameControlDisplayModes.clothItemPresetSelect)
            {
                _uiClothItemPresetSelector.UpdateGazeTouchDistances();
            }
            else if (GameControlUI.gameControlDisplayMode == GameControlDisplayModes.atomSelect)
            {
                _uiTargetAtomSelector.UpdateTASGazeTouchDistances();
            }

        }

        public void Update()
        {
            gridsCanvas.Update();

            if (GameControlUI.gridsActivated)
            {
                UpdateButtonDistances();

                // Supports button repositioning when in UIA Edit mode
                if (GameControlUI.gameControlDisplayMode == GameControlDisplayModes.edit) UpdateEditModeButtonPosition();

                // Enables or disable the Gaze Assist Select select box
                if (GazeAssistedSelectTool.gazeAssistActiveJSB.val && SuperController.singleton.gameMode == SuperController.GameMode.Edit && _gasButtonScrollListGO == null && GameControlUI.gameControlDisplayMode == GameControlDisplayModes.standard) GameControlUI.RefreshWristUIButtonGrid();
                if (_gasButtonScrollListGO != null && (!GazeAssistedSelectTool.gazeAssistActiveJSB.val || SuperController.singleton.gameMode == SuperController.GameMode.Play)) GameControlUI.RefreshWristUIButtonGrid();

                if (GameControlUI.gameControlDisplayMode == GameControlDisplayModes.standard || GameControlUI.gameControlDisplayMode == GameControlDisplayModes.edit)
                {

                    _uiMainButtonGridDisplay.UpdateUIAButtonGrid();

                    if (GameControlSettings.activateQuickLaunchBarJSB.val) _uiQLBButtonGridDisplay.UpdateUIAButtonGrid();

                    UpdateGASScrollList();
                }
                else if (GameControlUI.gameControlDisplayMode == GameControlDisplayModes.ace)
                {
                    string targetAtom = "";

                    if (_uiActiveClothingEditor.aceMode == ACEMode.original || _uiActiveClothingEditor.isGazeSelectedMode)
                    {
                        if (_uiActiveClothingEditor.gazeSelectedTargetType == LastViewedTargetType.lastViewedPerson) targetAtom = TargetControl.lastViewedPerson;
                        else if (_uiActiveClothingEditor.gazeSelectedTargetType == LastViewedTargetType.lastViewedMale) targetAtom = TargetControl.lastViewedMale;
                        else targetAtom = TargetControl.lastViewedFemale;
                    }
                    else
                    {
                        targetAtom = _uiActiveClothingEditor.personAtomNamesJSSC.val;
                    }


                    if (targetAtom == "" || !TargetControl.atomActiveClothingDictionary.ContainsKey(targetAtom)) _uiActiveClothingEditor.UpdateACE(targetAtom, null);
                    else _uiActiveClothingEditor.UpdateACE(targetAtom, TargetControl.atomActiveClothingDictionary[targetAtom]);
                }
                else if (GameControlUI.gameControlDisplayMode == GameControlDisplayModes.clothItemPresetSelect)
                {
                    _uiClothItemPresetSelector.UpdateUI();
                }
                else if (GameControlUI.gameControlDisplayMode == GameControlDisplayModes.atomSelect)
                {
                    _uiTargetAtomSelector.UpdateTAS();
                }

                if (GameControlUI.gameControlDisplayMode == GameControlDisplayModes.standard) UpdateGazeStatusTextField(GameControlUI._currentGrid);

            }


        }

        public static void UpdateButtonTextures(UIAButton button)
        {
            if (GameControlUI.gridsActivated)
            {
                if (button.parentGrid == _uiMainButtonGridDisplay._buttonGrid) _uiMainButtonGridDisplay.UpdateButtonTextures(button);
                else if (button.parentGrid == _uiQLBButtonGridDisplay._buttonGrid) _uiQLBButtonGridDisplay.UpdateButtonTextures(button);
            }

        }

        private static void UpdateGazeStatusTextField(UIAButtonGrid buttonGrid)
        {
            if (GameControlSettings.displayGazeSelectionsJSB.val && (buttonGrid.IsSelectionTargetButton() || (GameControlSettings.activateQuickLaunchBarJSB.val && UIAStorables.quickLaunchButtonGrid.IsSelectionTargetButton())))
            {
                string text = "";
                float height = UIAConsts._gazeStatusMaxHeight / 5f;
                _gazeStatusTextField.gameObject.SetActive(true);
                if (buttonGrid._gazeSelectionFemales || (GameControlSettings.activateQuickLaunchBarJSB.val && UIAStorables.quickLaunchButtonGrid._gazeSelectionFemales))
                {
                    if (text != "") text = text + "\n";
                    text = text + "Gaze Target Female: " + TargetControl.lastViewedFemale;
                    height = height + UIAConsts._gazeStatusMaxHeight / 5f;
                }
                if (buttonGrid._gazeSelectionMales || (GameControlSettings.activateQuickLaunchBarJSB.val && UIAStorables.quickLaunchButtonGrid._gazeSelectionMales))
                {
                    if (text != "") text = text + "\n";
                    text = text + "Gaze Target Male: " + TargetControl.lastViewedMale;
                    height = height + UIAConsts._gazeStatusMaxHeight / 5f;
                }
                if (buttonGrid._gazeSelectionPerson || (GameControlSettings.activateQuickLaunchBarJSB.val && UIAStorables.quickLaunchButtonGrid._gazeSelectionPerson))
                {
                    if (text != "") text = text + "\n";
                    text = text + "Gaze Target Person: " + TargetControl.lastViewedPerson;
                    height = height + UIAConsts._gazeStatusMaxHeight / 5f;
                }
                if (buttonGrid._gazeSelectionNonPerson || (GameControlSettings.activateQuickLaunchBarJSB.val && UIAStorables.quickLaunchButtonGrid._gazeSelectionNonPerson))
                {
                    if (text != "") text = text + "\n";
                    text = text + "Gaze Target Non-Person Atom: " + TargetControl.lastViewedNonPerson;
                    height = height + UIAConsts._gazeStatusMaxHeight / 5f;
                }
                if (buttonGrid._gazeSelectionAtoms || (GameControlSettings.activateQuickLaunchBarJSB.val && UIAStorables.quickLaunchButtonGrid._gazeSelectionAtoms))
                {
                    if (text != "") text = text + "\n";
                    text = text + "Gaze Target Atom: " + TargetControl.lastViewedAtom;
                    height = height + UIAConsts._gazeStatusMaxHeight / 5f;
                }
                if (buttonGrid._vamSelectionFemales || (GameControlSettings.activateQuickLaunchBarJSB.val && UIAStorables.quickLaunchButtonGrid._vamSelectionFemales))
                {
                    if (text != "") text = text + "\n";
                    text = text + "Selected Target Female: " + TargetControl.lastSelectedFemale;
                    height = height + UIAConsts._gazeStatusMaxHeight / 5f;
                }
                if (buttonGrid._vamSelectionMales || (GameControlSettings.activateQuickLaunchBarJSB.val && UIAStorables.quickLaunchButtonGrid._vamSelectionMales))
                {
                    if (text != "") text = text + "\n";
                    text = text + "Selected Target Male: " + TargetControl.lastSelectedMale;
                    height = height + UIAConsts._gazeStatusMaxHeight / 5f;
                }
                if (buttonGrid._vamSelectionPerson || (GameControlSettings.activateQuickLaunchBarJSB.val && UIAStorables.quickLaunchButtonGrid._vamSelectionPerson))
                {
                    if (text != "") text = text + "\n";
                    text = text + "Selected Target Person: " + TargetControl.lastSelectedPerson;
                    height = height + UIAConsts._gazeStatusMaxHeight / 5f;
                }
                if (buttonGrid._vamSelectionNonPerson || (GameControlSettings.activateQuickLaunchBarJSB.val && UIAStorables.quickLaunchButtonGrid._vamSelectionNonPerson))
                {
                    if (text != "") text = text + "\n";
                    text = text + "Selected Target Non-Person: " + TargetControl.lastSelectedNonPerson;
                    height = height + UIAConsts._gazeStatusMaxHeight / 5f;
                }
                if (buttonGrid._vamSelectionAtoms || (GameControlSettings.activateQuickLaunchBarJSB.val && UIAStorables.quickLaunchButtonGrid._vamSelectionAtoms))
                {
                    if (text != "") text = text + "\n";
                    text = text + "Selected Target Atom: " + TargetControl.lastSelectedAtom;
                    height = height + UIAConsts._gazeStatusMaxHeight / 5f;
                }
                foreach (CustomTargetGroupType cgt in CustomTargetGroupSettings.customTargetGroupTypeList)
                {
                    string cgtName = cgt.cgtNameJSS.val;
                    if ((buttonGrid._gazeSelectionCTGDic.ContainsKey(cgtName) && buttonGrid._gazeSelectionCTGDic[cgtName]) || (UIAStorables.quickLaunchButtonGrid._gazeSelectionCTGDic.ContainsKey(cgtName) && UIAStorables.quickLaunchButtonGrid._gazeSelectionCTGDic[cgtName]) && TargetControl.lastViewedCTGDic.ContainsKey(cgtName))
                    {
                        if (text != "") text = text + "\n";
                        text = text + "Gaze Target " + cgtName + ": " + TargetControl.lastViewedCTGDic[cgtName];
                        height = height + UIAConsts._gazeStatusMaxHeight / 5f;
                    }
                }
                _gazeStatusTextField.text = text;
                _gazeStatusTextField.height = height;
                _gazeStatusTextField.UItext.alignment = TextAnchor.MiddleLeft;
            }
            else if (_gazeStatusTextField != null) { _gazeStatusTextField.gameObject.SetActive(false); }

        }

        public float GetXYDistanceToGASScrollBar(Vector3 worldSpacePoint, float zThreshold)
        {
            float distance = 100000f;
            ScrollRect scrollRect = _gasButtonScrollListGO.GetComponent<ScrollRect>();
            Scrollbar sb = scrollRect.verticalScrollbar;
            RectTransform rt = sb.transform.GetComponent<RectTransform>();
            Vector3 localSpacePoint = rt.InverseTransformPoint(worldSpacePoint);

            if (localSpacePoint.z < zThreshold) distance = GeometryUtils.DistancePointToRectangle(new Vector2(localSpacePoint.x, localSpacePoint.y), rt.rect);

            return (distance);
        }
        public float GetGASScrollBarValue()
        {
            ScrollRect scrollRect = _gasButtonScrollListGO.GetComponent<ScrollRect>();
            Scrollbar sb = scrollRect.verticalScrollbar;
            return (sb.value);
        }
        public void OffsetGASScrollBarValueByYMovement(float initialSBValue, Vector3 initialWorldSpacePoint, Vector3 currentWorldSpacePoint)
        {
            ScrollRect scrollRect = _gasButtonScrollListGO.GetComponent<ScrollRect>();
            Scrollbar sb = scrollRect.verticalScrollbar;
            RectTransform rt = sb.transform.GetComponent<RectTransform>();
            Vector3 localInitialSpacePoint = rt.InverseTransformPoint(initialWorldSpacePoint);
            Vector3 localCurrentSpacePoint = rt.InverseTransformPoint(currentWorldSpacePoint);

            sb.value = Mathf.Clamp(initialSBValue + ((localCurrentSpacePoint.y - localInitialSpacePoint.y) / rt.rect.height), 0f, 1f);
        }
        public void CreateUIButtons()
        {
            if (GameControlUI.gameControlDisplayMode == GameControlDisplayModes.standard || GameControlUI.gameControlDisplayMode == GameControlDisplayModes.edit)
            {
                UIAButtonGrid buttonGrid = GameControlUI._currentGrid;

                Vector2 qlbSize = UIAStorables.quickLaunchButtonGrid.GetButtonDimensions();

                float verticalStartOffset = 30f;

                if (GameControlSettings.activateQuickLaunchBarJSB.val && PatreonFeatures.patreonContentEnabled)
                {
                    verticalStartOffset = PatreonFeatures.CreateQLBButtons(_uiQLBButtonGridDisplay, qlbSize);                   
                }

                _uiMainButtonGridDisplay.CreateUIButtons(buttonGrid, verticalStartOffset);
                _uiMainButtonGridDisplay.UpdateButtonTextures();

                float buttonsXMinWorld = _uiMainButtonGridDisplay.GetLeftEdgeXPosition();
                float buttonsXMaxWorld = _uiMainButtonGridDisplay.GetWidth() + buttonsXMinWorld;
                float buttonsYMaxWorld = _uiMainButtonGridDisplay.GetTopEdgeYPosition();
                float moveButtonsYMaxWorld = _uiMainButtonGridDisplay.GetTopEdgeYPosition(true);

                float buttonsWidth = _uiMainButtonGridDisplay.GetWidth();
                float buttonsHeight = _uiMainButtonGridDisplay.GetHeight();
                float scrollList1Width = UIAConsts._gasWidth;
                float scrollList1Height = Mathf.Max(buttonsHeight - UIAConsts._toggleButtonHeight, UIAConsts._gasMinHeight);

                if (GameControlUI.gameControlDisplayMode == GameControlDisplayModes.edit)
                {
                    int nextLeftGrid = GridsControlDisplay.GetNextLeftGridWithBlankButton();
                    int nextRightGrid = GridsControlDisplay.GetNextRightGridWithBlankButton();

                    if (nextLeftGrid != -1)
                    {
                        _moveUIDButtonToLeftGrid = gridsCanvas.CreateTextFieldInCanvas(buttonsXMinWorld + 45f, moveButtonsYMaxWorld, 0f, 90f, 40f);

                        _moveUIDButtonToLeftGrid.text = "MOVE TO\nGRID " + (nextLeftGrid + 1).ToString();
                        _moveUIDButtonToLeftGrid.UItext.fontSize = 12;
                        _moveUIDButtonToLeftGrid.UItext.alignment = TextAnchor.MiddleCenter;
                        _moveUIDButtonToLeftGrid.backgroundColor = Color.red;
                        _moveUIDButtonToLeftGrid.UItext.color = Color.white;
                        _moveUIDButtonToLeftGrid.gameObject.SetActive(false);
                    }
                    if (nextRightGrid != -1)
                    {
                        float xposition = buttonsXMaxWorld - 45f;
                        if (nextLeftGrid != -1 && xposition < buttonsXMinWorld + 90f) xposition = buttonsXMinWorld + 90f + 45f;
                        _moveUIDButtonToRightGrid = gridsCanvas.CreateTextFieldInCanvas(xposition, moveButtonsYMaxWorld, 0f, 90f, 40f);
                        _moveUIDButtonToRightGrid.text = "MOVE TO\nGRID " + (nextRightGrid + 1).ToString();
                        _moveUIDButtonToRightGrid.UItext.fontSize = 12;
                        _moveUIDButtonToRightGrid.UItext.alignment = TextAnchor.MiddleCenter;
                        _moveUIDButtonToRightGrid.backgroundColor = Color.red;
                        _moveUIDButtonToRightGrid.UItext.color = Color.white;
                        _moveUIDButtonToRightGrid.gameObject.SetActive(false);
                    }

                }

                if (GameControlUI.gameControlDisplayMode != GameControlDisplayModes.edit)
                {
                    _gazeStatusTextField = gridsCanvas.CreateTextFieldInCanvas(0f, buttonsYMaxWorld, 0f, Mathf.Max(buttonsWidth, UIAConsts._gazeStatusMinWidth), UIAConsts._gazeStatusMaxHeight);

                    _gazeStatusTextField.UItext.fontSize = 12;
                    _gazeStatusTextField.backgroundColor = new Color(173f / 255f, 216f / 255f, 230f / 255f, 1f);
                }

                if (GazeAssistedSelectTool.gazeAssistActiveJSB.val && SuperController.singleton.gameMode == SuperController.GameMode.Edit && GameControlUI.gameControlDisplayMode != GameControlDisplayModes.edit)
                {
                    bool scrollListLeftNotRight = false;
                    if (UIAGlobals.vrActive == false || VRSettings.vrHandControlJSEnum.val == VRHandControl.rightHand) scrollListLeftNotRight = true;

                    float listXoffset = (buttonsWidth / 2f) + (scrollList1Width / 2f);
                    float toggleXoffet = (buttonsWidth / 2f);
                    if (scrollListLeftNotRight)
                    {
                        listXoffset = -(buttonsWidth / 2f) - (scrollList1Width / 2f);
                        toggleXoffet = -(buttonsWidth / 2f) - (scrollList1Width);
                    }

                    _gasButtonScrollListGO = gridsCanvas.CreateScrollableButtonList(listXoffset, (scrollList1Height / 2f) + verticalStartOffset + UIAConsts._toggleButtonHeight, scrollList1Width, scrollList1Height, new Color(0.5f, 0.5f, 0.5f));
                    _gasHideControlNodesToggleGO = UITools.CreateToggle(gridsCanvas._canvas.transform, toggleXoffet, verticalStartOffset, scrollList1Width, UIAConsts._toggleButtonHeight, "Hide Inactive Targets", 11);

                    UIDynamicToggle uiToggle = _gasHideControlNodesToggleGO.transform.GetComponent<UIDynamicToggle>();
                    uiToggle.toggle.onValueChanged.AddListener((bool value) => { UserPreferences.singleton.hideInactiveTargets = value; });
                }
            }
            else if (GameControlUI.gameControlDisplayMode == GameControlDisplayModes.ace)
            {
                _uiActiveClothingEditor.CreateUI();
            }
            else if (GameControlUI.gameControlDisplayMode == GameControlDisplayModes.clothItemPresetSelect)
            {
                _uiClothItemPresetSelector.CreateUI();
            }
            else if (GameControlUI.gameControlDisplayMode == GameControlDisplayModes.atomSelect)
            {
                _uiTargetAtomSelector.CreateUI();
            }

        }
        public void DestroyUIButtons()
        {
            _uiMainButtonGridDisplay.DestroyUIButtons();
            _uiQLBButtonGridDisplay.DestroyUIButtons();

            if (_moveUIDButtonToLeftGrid != null) { GameObject.Destroy(_moveUIDButtonToLeftGrid.gameObject); }
            if (_moveUIDButtonToRightGrid != null) { GameObject.Destroy(_moveUIDButtonToRightGrid.gameObject); }

            if (_gazeStatusTextField != null) { GameObject.Destroy(_gazeStatusTextField.gameObject); }
            if (_gasButtonScrollListGO != null)
            {
                GameObject.Destroy(_gasButtonScrollListGO);
                GameObject.Destroy(_gasHideControlNodesToggleGO);
                _gasButtonScrollListGO = null;
                _gasHideControlNodesToggleGO = null;
                foreach (GameObject go in _gasScrollListButtonListGOs) GameObject.Destroy(go);
                _gasScrollListButtonListGOs.Clear();
            }
            _uiActiveClothingEditor.DestroyUI();
            _uiTargetAtomSelector.DestroyUI();
            _uiClothItemPresetSelector.DestoryUI();

        }


        public void UpdateGASScrollList()
        {
            if (_gasButtonScrollListGO != null)
            {
                Transform cameraT = SuperController.singleton.lookCamera.transform;
                Vector3 eyePosition = cameraT.position - (cameraT.up * 0.1f);
                Ray ray = new Ray(eyePosition, cameraT.forward);

                UIDynamicToggle uiToggle = _gasHideControlNodesToggleGO.transform.GetComponent<UIDynamicToggle>();
                uiToggle.toggle.isOn = UserPreferences.singleton.hideInactiveTargets;
                uiToggle.backgroundColor = new Color(0.8f, 0.8f, 0.8f);

                SetGASButtonCount(_gasScrollListButtonListGOs, _orderedGazeAssistSelectionTargets.Count);
                RectTransform scrollListRT = _gasButtonScrollListGO.GetComponent<RectTransform>();
                float checkRectMinY = scrollListRT.rect.yMin;
                float checkRectMaxY = scrollListRT.rect.yMax;


                for (int index = 0; index < _orderedGazeAssistSelectionTargets.Count; index++)
                {
                    UIDynamicButton selectTargetButton = _gasScrollListButtonListGOs[index].transform.GetComponent<UIDynamicButton>();
                    if (_orderedGazeAssistSelectionTargets[index].nodeName == "control") selectTargetButton.label = _orderedGazeAssistSelectionTargets[index].atomName;
                    else selectTargetButton.label = _orderedGazeAssistSelectionTargets[index].atomName + ":" + _orderedGazeAssistSelectionTargets[index].nodeName;
                    if (SuperController.singleton.GetSelectedAtom()?.name == _orderedGazeAssistSelectionTargets[index].atomName && SuperController.singleton.GetSelectedController()?.name == _orderedGazeAssistSelectionTargets[index].nodeName)
                    {
                        selectTargetButton.buttonColor = Color.yellow;
                    }
                    else selectTargetButton.buttonColor = new Color(0.8f, 0.8f, 0.8f);

                }
            }
        }
        public void UpdateGASButtonColor(int buttonIndex, Color color)
        {
            _gasScrollListButtonListGOs[buttonIndex].GetComponent<UIDynamicButton>().buttonColor = color;
        }
        private void SetGASButtonCount(List<GameObject> buttonListGOs, int buttonTargetCount)
        {
            if (buttonListGOs.Count > buttonTargetCount)
            {
                int buttonsToRemove = buttonListGOs.Count - buttonTargetCount;
                for (int counter = 1; counter <= buttonsToRemove; counter++)
                {
                    GameObject go = buttonListGOs[buttonListGOs.Count - 1];
                    GameObject.Destroy(go);
                    buttonListGOs.RemoveAt(buttonListGOs.Count - 1);
                }
                TargetControl.UpdateGazeTargets();
            }
            else if (buttonListGOs.Count < buttonTargetCount)
            {
                GameObject buttonListContentGO = _gasButtonScrollListGO.transform.Find("buttonListViewportGO/buttonListContentGO").gameObject;
                int buttonsToAdd = buttonTargetCount - buttonListGOs.Count;
                int intialButtonListGOCount = buttonListGOs.Count;
                for (int counter = 1; counter <= buttonsToAdd; counter++)
                {

                    GameObject newButtonGO = UITools.CreateUIButtonInParent(buttonListContentGO.transform);
                    RectTransform childRect = newButtonGO.GetComponent<RectTransform>();
                    childRect.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, 24f);
                    childRect.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Left, -10f, 200f);


                    buttonListGOs.Add(newButtonGO);
                    UIDynamicButton selectTargetButton = newButtonGO.transform.GetComponent<UIDynamicButton>();
                    selectTargetButton.buttonText.alignment = TextAnchor.MiddleLeft;
                    RectTransform rt = selectTargetButton.buttonText.gameObject.GetComponent<RectTransform>();
                    rt.anchoredPosition = new Vector2(0, 0);
                    rt.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Left, 20f, 400f);
                    rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, 24f);
                    selectTargetButton.buttonText.fontSize = 12;

                    int index = intialButtonListGOCount + counter - 1;
                    newButtonGO.SetActive(true);

                    selectTargetButton.button.onClick.AddListener(delegate
                    {
                        try
                        {
                            GazeAssistSelectButtonActivated(index);
                        }
                        catch (Exception e) { SuperController.LogError("Exception caught: " + e); }

                    });

                }
                TargetControl.UpdateGazeTargets();
            }
        }
        public void GazeAssistSelectButtonActivated(int index)
        {
            if (SuperController.singleton.GetSelectedAtom()?.name == _orderedGazeAssistSelectionTargets[index].atomName && SuperController.singleton.GetSelectedController()?.name == _orderedGazeAssistSelectionTargets[index].nodeName)
            {
                SuperController.singleton.SelectController(_orderedGazeAssistSelectionTargets[index].atomName, "control", false);
            }
            else SuperController.singleton.SelectController(_orderedGazeAssistSelectionTargets[index].atomName, _orderedGazeAssistSelectionTargets[index].nodeName, false);

        }
        public void UpdateEditModeButtonPosition()
        {
            _uiMainButtonGridDisplay.UpdateEditModeButtonPosition();
            _uiQLBButtonGridDisplay.UpdateEditModeButtonPosition();
            if (_uiMainButtonGridDisplay.IsButtonMoveActive())
            {
                if (_moveUIDButtonToLeftGrid != null) _moveUIDButtonToLeftGrid.gameObject.SetActive(true);
                if (_moveUIDButtonToRightGrid != null) _moveUIDButtonToRightGrid.gameObject.SetActive(true);

                if (GameControlUI.gameControlUIMode == GameControlUIModes.vrVAMUI) GridsControlDisplay.vamUILaunchButton.gameObject.SetActive(false);
                else GridsControlDisplay._mainButton.gameObject.SetActive(false);
                GridsControlDisplay._leftButton.gameObject.SetActive(false);
                GridsControlDisplay._rightButton.gameObject.SetActive(false);

            }
            else if (_uiQLBButtonGridDisplay.IsButtonMoveActive())
            {
                if (GameControlUI.gameControlUIMode == GameControlUIModes.vrVAMUI) GridsControlDisplay.vamUILaunchButton.gameObject.SetActive(false);
                else GridsControlDisplay._mainButton.gameObject.SetActive(false);
                GridsControlDisplay._leftButton.gameObject.SetActive(false);
                GridsControlDisplay._rightButton.gameObject.SetActive(false);
            }
            else
            {
                if (_moveUIDButtonToLeftGrid != null) _moveUIDButtonToLeftGrid.gameObject.SetActive(false);
                if (_moveUIDButtonToRightGrid != null) _moveUIDButtonToRightGrid.gameObject.SetActive(false);
                if (GameControlUI.gameControlUIMode == GameControlUIModes.vrVAMUI) GridsControlDisplay.vamUILaunchButton.gameObject.SetActive(true);
                else GridsControlDisplay._mainButton.gameObject.SetActive(true);
                GridsControlDisplay._leftButton.gameObject.SetActive(true);
                GridsControlDisplay._rightButton.gameObject.SetActive(true);
            }
        }

        public SingleGridDisplay GetAlternateBGDInMoveRange(Vector3 buttonWorldPosition, SingleGridDisplay uidBGD)
        {
            SingleGridDisplay alternateUIDbgd = null;
            if (GameControlSettings.activateQuickLaunchBarJSB.val)
            {
                float verticalOffset = _uiQLBButtonGridDisplay.GetHeight() + 10f;
                Vector3 localPosition = gridsCanvas._canvas.transform.InverseTransformPoint(buttonWorldPosition);
                if (uidBGD == _uiQLBButtonGridDisplay)
                {
                    if (_uiMainButtonGridDisplay._buttonGrid.ContainsBlankButtons() && localPosition.y > verticalOffset + 5f)
                    {
                        return (_uiMainButtonGridDisplay);
                    }
                }
                else if (_uiQLBButtonGridDisplay._buttonGrid.ContainsBlankButtons())
                {
                    if (localPosition.y < verticalOffset - 5f)
                    {
                        return (_uiQLBButtonGridDisplay);
                    }
                }
            }
            return (alternateUIDbgd);
        }

        public void CheckMovingButtonInLeftMoveRange(Vector3 buttonWorldPosition, float buttonDistanceOffsetFromCanvas)
        {
            _moveUIDButtonToLeftGridInRange = false;
            if (_moveUIDButtonToLeftGrid != null && _moveUIDButtonToLeftGrid.gameObject.activeSelf)
            {
                Vector3[] corners = new Vector3[4];
                _moveUIDButtonToLeftGrid.UItext.rectTransform.GetWorldCorners(corners);

                Vector3 buttonCentre = corners[0] + ((corners[2] - corners[0]) * 0.5f);

                float distance = Vector3.Distance(buttonWorldPosition, buttonCentre);
                if (distance < 0.015f + buttonDistanceOffsetFromCanvas) _moveUIDButtonToLeftGridInRange = true;

                if (_moveUIDButtonToLeftGridInRange)
                {
                    Color cl = new Color(0.4f, 0f, 0f);
                    _moveUIDButtonToLeftGrid.backgroundColor = cl;
                    _moveUIDButtonToLeftGrid.UItext.color = Color.white;
                }
                else
                {
                    _moveUIDButtonToLeftGrid.backgroundColor = Color.red;
                    _moveUIDButtonToLeftGrid.UItext.color = Color.white;
                }

            }
        }
        public void CheckMovingButtonInRightMoveRange(Vector3 buttonWorldPosition, float buttonDistanceOffsetFromCanvas)
        {
            _moveUIDButtonToRightGridInRange = false;
            if (_moveUIDButtonToRightGrid != null && _moveUIDButtonToRightGrid.gameObject.activeSelf)
            {
                Vector3[] corners = new Vector3[4];
                _moveUIDButtonToRightGrid.UItext.rectTransform.GetWorldCorners(corners);

                Vector3 buttonCentre = corners[0] + ((corners[2] - corners[0]) * 0.5f);

                float distance = Vector3.Distance(buttonWorldPosition, buttonCentre);
                if (distance < 0.015f + buttonDistanceOffsetFromCanvas) _moveUIDButtonToRightGridInRange = true;
                if (_moveUIDButtonToRightGridInRange)
                {
                    Color cl = new Color(0.4f, 0f, 0f);
                    _moveUIDButtonToRightGrid.backgroundColor = cl;
                    _moveUIDButtonToRightGrid.UItext.color = Color.white;
                }
                else
                {
                    _moveUIDButtonToRightGrid.backgroundColor = Color.red;
                    _moveUIDButtonToRightGrid.UItext.color = Color.white;
                }
            }
        }
    }

    public class SingleGridDisplay
    {
        private List<UIDynamicButton> _uiButtonList;
        private GridsCanvas _uiButtonCanvas;
        public UIAButtonGrid _buttonGrid;

        private float _verticalStartOffset;
        private float _horizontalStartOffset;
        private Vector2 _buttonDimensions;
        private int _columns;
        private int _rows;
        private bool _buttonMoveActive = false;
        public int _buttonMoveIndex = -1;
        public float _buttonMoveStartTime = 0f;
        public string _buttonMoveController = "";

        public static void ButtonActivated(int buttonIndex, UIAButtonGrid buttonGrid, int vrHand = LeftRight.neither, bool leapActivated = false)
        {
            try
            {
                UIAButton button = buttonGrid.buttonList[buttonIndex];
                if (RelativePositionComponentUITab.testSpawnModeActive) RelativePositionComponentUITab.DeactivateSpawnTest();

                if (vrHand == LeftRight.neither)
                {
                    if (SuperController.singleton.GetRightSelect()) vrHand = LeftRight.right;
                    if (SuperController.singleton.GetLeftSelect()) vrHand = LeftRight.left;
                }

                if (SuperController.singleton.mainHUD == null) UIAButton.hudOnPreButtonAction = false;
                else UIAButton.hudOnPreButtonAction = SuperController.singleton.mainHUD.gameObject.activeInHierarchy;

                if (GameControlUI.gameControlDisplayMode != GameControlDisplayModes.edit) button.ActionInit(vrHand, leapActivated);

            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }

        public bool IsButtonMoveActive()
        {
            return (_buttonMoveActive && (Time.time - _buttonMoveStartTime) > 0.4f);
        }

        public SingleGridDisplay(GridsCanvas uiButtonCanvas)
        {
            _uiButtonList = new List<UIDynamicButton>();
            _uiButtonCanvas = uiButtonCanvas;
        }
        public void RefreshButtonPositions()
        {
            int buttonNum = 0;
            for (int row = 0; row < _rows; row++)
            {
                for (int col = 0; col < _columns; col++)
                {
                    UIDynamicButton gridButton = _uiButtonList[buttonNum];

                    gridButton.button.transform.localPosition = new Vector3((-_horizontalStartOffset + (col * _buttonDimensions.x)), (_verticalStartOffset + ((row + 0.5f) * _buttonDimensions.y)), 0f);

                    buttonNum++;
                }
            }
        }
        public void CreateUIButtons(UIAButtonGrid buttonGrid, float verticalStartOffset)
        {
            _buttonGrid = buttonGrid;
            _buttonDimensions = buttonGrid.GetButtonDimensions();

            Vector2 gridSize = PatreonFeatures.GridSize(buttonGrid);
            _rows = (int) gridSize.y;
            _columns = (int) gridSize.x;

            _verticalStartOffset = verticalStartOffset;
            _horizontalStartOffset = _buttonDimensions.x * ((_columns - 1) / 2f);

            _uiButtonList.Clear();

            int buttonRef = 0;
            for (int row = 0; row < _rows; row++)
            {
                for (int col = 0; col < _columns; col++)
                {
                    buttonRef++;
                    UIDynamicButton gridButton = _uiButtonCanvas.CreateUIButtonInCanvas(buttonRef.ToString(), (-_horizontalStartOffset + (col * _buttonDimensions.x)), (_verticalStartOffset + ((row + 0.5f) * _buttonDimensions.y)), 0f, _buttonDimensions.x, _buttonDimensions.y);

                    int buttonIndex = buttonRef - 1;
                    gridButton.button.onClick.AddListener(delegate {
                        ButtonActivated(buttonIndex, buttonGrid);
                    });

                    EventTrigger trigger = gridButton.button.gameObject.AddComponent<EventTrigger>();
                    var pointerDown = new EventTrigger.Entry();
                    pointerDown.eventID = EventTriggerType.PointerDown;
                    pointerDown.callback.AddListener((e) => GridButtonClickDown(buttonIndex));
                    trigger.triggers.Add(pointerDown);

                    var pointerUp = new EventTrigger.Entry();
                    pointerUp.eventID = EventTriggerType.PointerUp;

                    pointerUp.callback.AddListener((e) => GridButtonClickUp(buttonIndex));
                    trigger.triggers.Add(pointerUp);

                    RectTransform rt = gridButton.buttonText.gameObject.GetComponent<RectTransform>();
                    rt.anchoredPosition = new Vector2(0, 0);
                    rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, _buttonDimensions.x * 0.8f);
                    rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, _buttonDimensions.y * 0.8f);

                    _uiButtonList.Add(gridButton);
                }
            }
        }
        public void DestroyUIButtons()
        {
            foreach (UIDynamicButton uiButton in _uiButtonList)
            {
                if (uiButton != null)
                {
                    uiButton.button.onClick.RemoveAllListeners();

                    GameObject.Destroy(uiButton.gameObject);
                }
            }
        }
        public float GetLeftEdgeXPosition()
        {
            return (-_horizontalStartOffset - (_buttonDimensions.x / 2f));
        }
        public float GetTopEdgeYPosition(bool editMode = false)
        {
            int buttonRows = _buttonGrid._maxButtonRow;
            if (editMode) buttonRows = _buttonGrid.gridRows;
            return (_verticalStartOffset + (buttonRows * _buttonDimensions.y));
        }
        public float GetWidth()
        {
            return (_columns * _buttonDimensions.x);
        }
        public float GetHeight()
        {
            return (_rows * _buttonDimensions.y);
        }

        public void UpdateButtonTextures(UIAButton button)
        {
            UIDynamicButton uiButton = _uiButtonList[button.buttonGridIndex];
            Texture2D thumbnailTexture;

            if (!button.ContainsAllBlankOperations() && button.buttonTexture != null)
            {
                thumbnailTexture = button.buttonTexture;
            }
            else
            {
                int buttonState = button.GetButtonState();
                if (buttonState == ButtonState.inactive) thumbnailTexture = UIAStorables.defaultButtonSkin.buttonSkinOnParams.buttonTexture;
                else thumbnailTexture = UIAStorables.defaultButtonSkin.buttonSkinOffParams.buttonTexture; ;

            }

            if (thumbnailTexture != null)
            {
                Sprite mySprite = Sprite.Create(thumbnailTexture, new Rect(0.0f, 0.0f, thumbnailTexture.width, thumbnailTexture.height), new Vector2(0.5f, 0.5f), 100.0f);
                uiButton.buttonImage.sprite = mySprite;
            }
            else
            {
                
                uiButton.buttonImage.sprite = UITools.defaultButtonSprite;
            }

        }

        public void UpdateButtonTextures()
        {

            int buttonCount = _uiButtonList.Count;
            List<string> test = new List<string>();
            test.Add("Temp");
            PatreonFeatures.GridLayoutFeature(test);
            if (!PatreonFeatures.patreonContentEnabled || test == null)
            {
                buttonCount = Mathf.Min(buttonCount, 16);
            }

            for (int buttonRef = 0; buttonRef < buttonCount; buttonRef++)
            {
                UpdateButtonTextures(_buttonGrid.buttonList[buttonRef]);
            }
        }

        public void UpdateButtonColor(int buttonIndex, Color color)
        {
            if (buttonIndex < _uiButtonList.Count && buttonIndex >= 0)
            {
                _uiButtonList[buttonIndex].buttonColor = color;
            }

        }
        public void UpdateButtonDistances()
        {

            for (int buttonRef = 0; buttonRef < _uiButtonList.Count; buttonRef++)
            {
                UIDynamicButton uiButton = _uiButtonList[buttonRef];
                UIAButton button = _buttonGrid.buttonList[buttonRef];
                if (!button.ContainsAllBlankOperations() || GameControlUI.gameControlDisplayMode == GameControlDisplayModes.edit)
                {
                    GameControlUI.SetNearestGazeAndLeapButtons(uiButton, _buttonGrid, buttonRef);
                }
            }
        }

        public void UpdateUIAButtonGrid()
        {
            int buttonCount = _uiButtonList.Count;
            Transform cameraT = SuperController.singleton.lookCamera.transform;
            Vector3 eyePosition = cameraT.position - (cameraT.up * 0.1f);
            Ray ray = new Ray(eyePosition, cameraT.forward);

            for (int buttonRef = 0; buttonRef < buttonCount; buttonRef++)
            {
                UIDynamicButton uiButton = _uiButtonList[buttonRef];
                UIAButton button = _buttonGrid.buttonList[buttonRef];
                ButtonSkinComponent buttonSkin = button.buttonSkinComponent;

                int buttonState = button.GetButtonState();

                if (button.ContainsAllBlankOperations() && GameControlUI.gameControlDisplayMode != GameControlDisplayModes.edit)
                {
                    uiButton.gameObject.SetActive(false);
                }
                else
                {
                    uiButton.gameObject.SetActive(true);

                    string targetFont = "Arial";
                    if (buttonSkin == null || buttonSkin.buttonFontDefault.val) targetFont = UIAStorables.defaultButtonSkin.buttonFont.val;
                    else { targetFont = buttonSkin.buttonFont.val; }

                    Font selectedFont = (Font)Resources.GetBuiltinResource(typeof(Font), "Arial.ttf");
                    foreach (Font font in Resources.FindObjectsOfTypeAll(typeof(Font)) as Font[])
                    {
                        if (font.name == targetFont) selectedFont = font;
                    }
                    uiButton.buttonText.font = selectedFont;

                    HSVColor hsvTextColor = new HSVColor();
                    HSVColor hsvButtonColor = new HSVColor();
                    if (button.ContainsAllBlankOperations()) uiButton.label = "";
                    else if (buttonState == 1)
                    {
                        if (button.autoLabelJSBool.val && button.buttonOperations.Count == 1) uiButton.label = button.GetAutoLabel();
                        else uiButton.label = button.GetManualLabel();
                        if (buttonSkin == null || buttonSkin.buttonSkinOnParams.textColorUseDefault.val) hsvTextColor = UIAStorables.defaultButtonSkin.buttonSkinOnParams.textColor.val;
                        else hsvTextColor = buttonSkin.buttonSkinOnParams.textColor.val;
                        if (buttonSkin == null || buttonSkin.buttonSkinOnParams.buttonColorUseDefault.val) hsvButtonColor = UIAStorables.defaultButtonSkin.buttonSkinOnParams.buttonColor.val;
                        else hsvButtonColor = buttonSkin.buttonSkinOnParams.buttonColor.val;
                    }
                    else if (buttonState == 2)
                    {
                        if (button.autoLabelJSBool.val && button.buttonOperations.Count == 1) uiButton.label = button.GetAutoLabel();
                        else uiButton.label = button.GetManualLabel();
                        if (buttonSkin == null || buttonSkin.buttonSkinOffParams.textColorUseDefault.val) hsvTextColor = UIAStorables.defaultButtonSkin.buttonSkinOffParams.textColor.val;
                        else hsvTextColor = buttonSkin.buttonSkinOffParams.textColor.val;
                        if (buttonSkin == null || buttonSkin.buttonSkinOffParams.buttonColorUseDefault.val) hsvButtonColor = UIAStorables.defaultButtonSkin.buttonSkinOffParams.buttonColor.val;
                        else hsvButtonColor = buttonSkin.buttonSkinOffParams.buttonColor.val;
                    }
                    if (buttonSkin == null || buttonSkin.fontSizeDefault.val)
                    {
                        if (_buttonGrid._buttonSize == "Micro")
                        {
                            uiButton.buttonText.resizeTextMaxSize = (int)UIAStorables.defaultButtonSkin.defaultMicroFontSize.maxFontSize.val * 2;
                            uiButton.buttonText.resizeTextMinSize = (int)UIAStorables.defaultButtonSkin.defaultMicroFontSize.minFontSize.val * 2;
                        }
                        else if (_buttonGrid._buttonSize == "Mini")
                        {
                            uiButton.buttonText.resizeTextMaxSize = (int)UIAStorables.defaultButtonSkin.defaultMiniFontSize.maxFontSize.val * 2;
                            uiButton.buttonText.resizeTextMinSize = (int)UIAStorables.defaultButtonSkin.defaultMiniFontSize.minFontSize.val * 2;
                        }
                        else if (_buttonGrid._buttonSize == "Small")
                        {
                            uiButton.buttonText.resizeTextMaxSize = (int)UIAStorables.defaultButtonSkin.defaultSmallFontSize.maxFontSize.val * 2;
                            uiButton.buttonText.resizeTextMinSize = (int)UIAStorables.defaultButtonSkin.defaultSmallFontSize.minFontSize.val * 2;
                        }
                        else if (_buttonGrid._buttonSize == "Medium")
                        {
                            uiButton.buttonText.resizeTextMaxSize = (int)UIAStorables.defaultButtonSkin.defaultMediumFontSize.maxFontSize.val * 2;
                            uiButton.buttonText.resizeTextMinSize = (int)UIAStorables.defaultButtonSkin.defaultMediumFontSize.minFontSize.val * 2;
                        }
                        else if (_buttonGrid._buttonSize == "Large")
                        {
                            uiButton.buttonText.resizeTextMaxSize = (int)UIAStorables.defaultButtonSkin.defaultLargeFontSize.minFontSize.val * 2;
                            uiButton.buttonText.resizeTextMinSize = (int)UIAStorables.defaultButtonSkin.defaultLargeFontSize.minFontSize.val * 2;
                        }
                    }
                    else
                    {
                        uiButton.buttonText.resizeTextMaxSize = (int)buttonSkin.fontSize.maxFontSize.val * 2;
                        uiButton.buttonText.resizeTextMinSize = (int)buttonSkin.fontSize.minFontSize.val * 2;
                    }
                    uiButton.buttonText.resizeTextForBestFit = true;
                    uiButton.buttonText.color = Color.HSVToRGB(hsvTextColor.H, hsvTextColor.S, hsvTextColor.V);
                    if (button.ContainsAllBlankOperations()) uiButton.buttonColor = new Color(0.85f, 0.85f, 0.85f);
                    else uiButton.buttonColor = Color.HSVToRGB(hsvButtonColor.H, hsvButtonColor.S, hsvButtonColor.V);

                    string textAlignment = "";
                    if (buttonSkin != null) textAlignment = buttonSkin.textAlignment.val;
                    if (buttonSkin == null || buttonSkin.textAlignmentDefault.val) { textAlignment = UIAStorables.defaultButtonSkin.textAlignment.val; }
                    if (textAlignment == "LowerCenter") { uiButton.buttonText.alignment = TextAnchor.LowerCenter; }
                    else if (textAlignment == "MiddleCenter") { uiButton.buttonText.alignment = TextAnchor.MiddleCenter; }
                    else if (textAlignment == "UpperCenter") { uiButton.buttonText.alignment = TextAnchor.UpperCenter; }
                    else if (textAlignment == "LowerLeft") { uiButton.buttonText.alignment = TextAnchor.LowerLeft; }
                    else if (textAlignment == "MiddleLeft") { uiButton.buttonText.alignment = TextAnchor.MiddleLeft; }
                    else if (textAlignment == "UpperLeft") { uiButton.buttonText.alignment = TextAnchor.UpperLeft; }
                    else if (textAlignment == "LowerRight") { uiButton.buttonText.alignment = TextAnchor.LowerRight; }
                    else if (textAlignment == "MiddleRight") { uiButton.buttonText.alignment = TextAnchor.MiddleRight; }
                    else if (textAlignment == "UpperRight") { uiButton.buttonText.alignment = TextAnchor.UpperRight; }
                }
            }
        }
        public void GridButtonClickDown(int buttonIndex)
        {
            try
            {
                if (GameControlUI.gameControlDisplayMode == GameControlDisplayModes.edit)
                {
                    _buttonMoveIndex = buttonIndex;
                    _buttonMoveActive = true;
                    _buttonMoveStartTime = Time.time;
                    if (_buttonGrid == UIAStorables.quickLaunchButtonGrid)
                    {
                        UIAPluginUI.TopTabSelected(TopUITabs.buttonSetup);
                        ButtonSetupPluginUI.gridSelectorJSSC.val = "QL";
                    }
                    else
                    {
                        UIAPluginUI.TopTabSelected(TopUITabs.buttonSetup);
                        ButtonSetupPluginUI.gridSelectorJSSC.val = _buttonGrid.gridRef;
                    }
                    ButtonSetupPluginUI.buttonSelectorJSSC.val = _buttonGrid.GetButtonRCRef(buttonIndex);
                    UIAPluginUI.buttonSetupUI.RefreshGroupUI(LeftRight.both, 1);

                    SuperController.singleton.ShowMainHUD();
                    SuperController.singleton.activeUI = SuperController.ActiveUI.MainMenu;

                    SuperController.singleton.mainMenuTabSelector.SetActiveTab("TabSessionPlugins");
                    IEnumerator enumerator1 = SuperController.singleton.mainMenuTabSelector.transform.GetEnumerator();
                    while (enumerator1.MoveNext())
                    {
                        UITab component = ((Component)enumerator1.Current).GetComponent<UITab>();
                        if ((UnityEngine.Object)component != (UnityEngine.Object)null)
                        {
                            foreach (var scriptController in component.GetComponentsInChildren<MVRScriptControllerUI>())
                            {
                                if (scriptController.label.text == UIAGlobals.mvrScript.name)
                                {
                                    scriptController.openUIButton?.onClick?.Invoke();
                                    break;
                                }
                            }
                        }
                    }

                    if (UIAGlobals.vrActive && _buttonMoveController == "")
                    {
                        if (SuperController.singleton.GetRightSelect() && GridsDisplay._uiMainButtonGridDisplay._buttonMoveController != "RIGHT" && GridsDisplay._uiQLBButtonGridDisplay._buttonMoveController != "RIGHT")
                        {
                            _buttonMoveController = "RIGHT";
                        }
                        else if (SuperController.singleton.GetLeftSelect())
                        {
                            _buttonMoveController = "LEFT";
                        }
                    }
                }

            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public void GridButtonClickUp(int buttonIndex)
        {
            try
            {
                int nextGridIndex = -1;
                int nextGridBlankButtonIndex = -1;
                if (_buttonMoveActive && GameControlUI.gridsDisplay._moveUIDButtonToLeftGridInRange)
                {
                    nextGridIndex = GridsControlDisplay.GetNextLeftGridWithBlankButton();
                    nextGridBlankButtonIndex = UIAStorables.buttonGridsList[nextGridIndex].GetFirstBlankButtonIndex();

                }
                else if (_buttonMoveActive && GameControlUI.gridsDisplay._moveUIDButtonToRightGridInRange)
                {
                    nextGridIndex = GridsControlDisplay.GetNextRightGridWithBlankButton();
                    nextGridBlankButtonIndex = UIAStorables.buttonGridsList[nextGridIndex].GetFirstBlankButtonIndex();
                }          

                if (nextGridIndex > -1)
                {
                    UIAStorables.buttonGridsList[nextGridIndex].buttonList[nextGridBlankButtonIndex] = _buttonGrid.buttonList[buttonIndex];
                    _buttonGrid.buttonList[buttonIndex].parentGrid = UIAStorables.buttonGridsList[nextGridIndex];
                    UIAStorables.buttonGridsList[nextGridIndex].RecalcGazeSelections();
                    _buttonGrid.buttonList[buttonIndex] = new UIAButton(_buttonGrid);
                    _buttonGrid.RecalcGazeSelections();
                }
                _buttonMoveActive = false;
                _buttonMoveController = "";
                GameControlUI.gridsDisplay._moveUIDButtonToLeftGridInRange = false;
                GameControlUI.gridsDisplay._moveUIDButtonToRightGridInRange = false;

                if (GameControlUI.gameControlDisplayMode == GameControlDisplayModes.edit) GameControlUI.RefreshWristUIButtonGrid();
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public void UpdateEditModeButtonPosition()
        {
            if (IsButtonMoveActive() && _buttonMoveIndex >= 0)
            {
                UIDynamicButton uiButton = _uiButtonList[_buttonMoveIndex];
                Vector3 worldPosition = new Vector3();
                float buttonDistanceOffsetFromCanvas = 0f;

                if (!UIAGlobals.vrActive)
                {
                    Vector2 pos;
                    RectTransformUtility.ScreenPointToLocalPointInRectangle(_uiButtonCanvas._canvas.transform as RectTransform, Input.mousePosition, _uiButtonCanvas._canvas.worldCamera, out pos);
                    worldPosition = _uiButtonCanvas._canvas.transform.TransformPoint(pos);
                }
                else
                {
                    Plane plane = new Plane(_uiButtonCanvas._canvas.transform.forward, _uiButtonCanvas._canvas.transform.TransformPoint(new Vector3(0f, 0f, 0f)));

                    Transform lineLeft = SuperController.singleton.leftControllerCamera.transform;
                    Transform lineRight = SuperController.singleton.rightControllerCamera.transform;

                    Ray ray;

                    if (_buttonMoveController == "LEFT") ray = new Ray(lineLeft.transform.position, lineLeft.transform.rotation * Vector3.forward);
                    else ray = new Ray(lineRight.transform.position, lineRight.transform.rotation * Vector3.forward);

                    float distanceA;

                    if (plane.Raycast(ray, out distanceA))
                    {
                        worldPosition = ray.GetPoint(distanceA * 0.9f);
                        buttonDistanceOffsetFromCanvas = Vector3.Distance(worldPosition, ray.GetPoint(distanceA));
                    }
                }
                uiButton.transform.position = worldPosition;
                uiButton.transform.SetAsLastSibling();
                int nearestButtonIndex = GetUIAEditNearestGridButtonIndex(worldPosition);

                if (nearestButtonIndex != _buttonMoveIndex)
                {
                    UIDynamicButton tempButton;
                    UIAButton tempUIAButton;

                    tempButton = _uiButtonList[_buttonMoveIndex];
                    tempUIAButton = _buttonGrid.buttonList[_buttonMoveIndex];

                    ButtonShift(_buttonMoveIndex, nearestButtonIndex);

                    _uiButtonList[nearestButtonIndex] = tempButton;
                    _buttonGrid.buttonList[nearestButtonIndex] = tempUIAButton;
                    //                    _buttonGrid._buttonList[nearestButtonIndex]._buttonRef = nearestButtonIndex + 1;
                    EventTrigger trigger = tempButton.button.gameObject.GetComponent<EventTrigger>();
                    List<EventTrigger.Entry> triggerEntryList = trigger.triggers;
                    foreach (EventTrigger.Entry triggerEntry in triggerEntryList)
                    {
                        if (triggerEntry.eventID == EventTriggerType.PointerUp)
                        {
                            triggerEntry.callback.RemoveAllListeners();
                            triggerEntry.callback.AddListener((e) => GridButtonClickUp(nearestButtonIndex));
                        }
                    }


                    _buttonMoveIndex = nearestButtonIndex;

                    if (_buttonGrid == UIAStorables.quickLaunchButtonGrid)
                    {
                        UIAPluginUI.TopTabSelected(TopUITabs.buttonSetup);
                        ButtonSetupPluginUI.gridSelectorJSSC.val = "QL";
                    }
                    else
                    {
                        UIAPluginUI.TopTabSelected(TopUITabs.buttonSetup);
                        ButtonSetupPluginUI.gridSelectorJSSC.val = _buttonGrid.gridRef;
                    }

                    ButtonSetupPluginUI.buttonSelectorJSSC.val = _buttonGrid.GetButtonRCRef(nearestButtonIndex);
                    UIAPluginUI.buttonSetupUI.RefreshGroupUI(LeftRight.both, 1);

                    RefreshButtonPositions();
                }
                SingleGridDisplay alternateBGD = GameControlUI.gridsDisplay.GetAlternateBGDInMoveRange(worldPosition, this);

                if (alternateBGD != null && alternateBGD._buttonGrid.ContainsBlankButtons())
                {
                    alternateBGD.InsertButtonFromBGD(_buttonGrid.buttonList[_buttonMoveIndex], _uiButtonList[_buttonMoveIndex], worldPosition, _buttonMoveStartTime, _buttonMoveController);
                    _buttonMoveActive = false;

                    _buttonGrid.buttonList[_buttonMoveIndex] = new UIAButton(_buttonGrid);
                    _uiButtonList[_buttonMoveIndex] = null;
                    DestroyUIButtons();
                    CreateUIButtons(_buttonGrid, _verticalStartOffset);
                    UpdateButtonTextures();
                }
                else
                {
                    GameControlUI.gridsDisplay.CheckMovingButtonInLeftMoveRange(worldPosition, buttonDistanceOffsetFromCanvas);
                    GameControlUI.gridsDisplay.CheckMovingButtonInRightMoveRange(worldPosition, buttonDistanceOffsetFromCanvas);
                }
            }

        }
        public void InsertButtonFromBGD(UIAButton button, UIDynamicButton uidButton, Vector3 worldPosition, float startTime, string controller)
        {
            int nearestIndex = GetUIAEditNearestGridButtonIndex(worldPosition);
            int higherBlankButtonIndex = -1;
            int lowerBlankButtonIndex = -1;
            int nearestBlankButtonIndex = -1;
            for (int index = nearestIndex; index < _buttonGrid.buttonCount; index++)
            {
                if (_buttonGrid.buttonList[index].ContainsAllBlankOperations())
                {
                    higherBlankButtonIndex = index;
                    break;
                }
            }
            for (int index = nearestIndex; index >= 0; index--)
            {
                if (_buttonGrid.buttonList[index].ContainsAllBlankOperations())
                {
                    lowerBlankButtonIndex = index;
                    break;
                }
            }

            if (lowerBlankButtonIndex != -1 && (higherBlankButtonIndex == -1 || (higherBlankButtonIndex - nearestIndex) > (nearestIndex - lowerBlankButtonIndex))) nearestBlankButtonIndex = lowerBlankButtonIndex;
            else nearestBlankButtonIndex = higherBlankButtonIndex;

            UIDynamicButton nearestBlankUIDButton = _uiButtonList[nearestBlankButtonIndex];
            nearestBlankUIDButton.button.onClick.RemoveAllListeners();
            GameObject.Destroy(nearestBlankUIDButton.gameObject);

            ButtonShift(nearestBlankButtonIndex, nearestIndex);

            _uiButtonList[nearestIndex] = uidButton;
            _buttonGrid.buttonList[nearestIndex] = button;
            button.parentGrid = _buttonGrid;

            _buttonMoveIndex = nearestIndex;
            _buttonMoveActive = true;
            _buttonMoveController = controller;
            _buttonMoveStartTime = startTime;

            EventTrigger trigger = uidButton.button.gameObject.GetComponent<EventTrigger>();
            List<EventTrigger.Entry> triggerEntryList = trigger.triggers;
            foreach (EventTrigger.Entry triggerEntry in triggerEntryList)
            {
                if (triggerEntry.eventID == EventTriggerType.PointerUp)
                {
                    triggerEntry.callback.RemoveAllListeners();
                    triggerEntry.callback.AddListener((e) => GridButtonClickUp(nearestIndex));
                }
            }
            uidButton.button.GetComponent<RectTransform>().SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, _buttonDimensions.x);
            uidButton.button.GetComponent<RectTransform>().SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, _buttonDimensions.y);

            RectTransform rt = uidButton.buttonText.gameObject.GetComponent<RectTransform>();
            rt.anchoredPosition = new Vector2(0, 0);
            rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, _buttonDimensions.x * 0.8f);
            rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, _buttonDimensions.y * 0.8f);

            if (_buttonGrid == UIAStorables.quickLaunchButtonGrid)
            {
                UIAPluginUI.TopTabSelected(TopUITabs.buttonSetup);
                ButtonSetupPluginUI.gridSelectorJSSC.val = "QL";
            }
            else
            {
                UIAPluginUI.TopTabSelected(TopUITabs.buttonSetup);
                ButtonSetupPluginUI.gridSelectorJSSC.val = _buttonGrid.gridRef;
            }

            ButtonSetupPluginUI.buttonSelectorJSSC.val = _buttonGrid.GetButtonRCRef(nearestIndex);
            UIAPluginUI.buttonSetupUI.RefreshGroupUI(LeftRight.both, 1);

            RefreshButtonPositions();
        }
        public void ButtonShift(int startIndex, int endIndex)
        {
            int increment = 1;
            if (endIndex < startIndex) increment = -1;

            for (int index = startIndex; index != endIndex; index += increment)
            {
                _uiButtonList[index] = _uiButtonList[index + increment];
                _buttonGrid.buttonList[index] = _buttonGrid.buttonList[index + increment];
                //               _buttonGrid._buttonList[index]._buttonRef = index + 1;
            }
        }
        private int GetUIAEditNearestGridButtonIndex(Vector3 worldPosition)
        {
            Vector3 localPosition = _uiButtonCanvas._canvas.transform.InverseTransformPoint(worldPosition);

            int xPos = (int)Math.Round((localPosition.x + _horizontalStartOffset) / _buttonDimensions.x);
            int yPos = (int)Math.Round((localPosition.y - _verticalStartOffset - 20f) / _buttonDimensions.y);
            int xPosLimited = Math.Max(0, Math.Min(xPos, _columns - 1));
            int yPosLimited = Math.Max(0, Math.Min(yPos, _rows - 1));
            return ((yPosLimited * _columns) + xPosLimited);
        }

    }


    public class TargetAtomSelectorDisplay
    {
        public class UIDTargetAtomSelectorItem
        {
            private GridsCanvas _uiButtonCanvas;
            public GameObject _tasButtonGO;

            public UIDynamicButton _tasSelectButton;

            public UIDTargetAtomSelectorItem(GridsCanvas uiButtonCanvas, GameObject tasScrollListGO, string atomName, UIAButton.UserAtomSelectedCallback buttonCallback)
            {
                try
                {
                    _uiButtonCanvas = uiButtonCanvas;

                    float horizontalOffset = 5f;

                    if (tasScrollListGO != null)
                    {
                        GameObject buttonListContentGO = tasScrollListGO.transform.Find("buttonListViewportGO/buttonListContentGO").gameObject;

                        _tasButtonGO = UITools.CreateUIButtonInParent(buttonListContentGO.transform);
                        UITools.SetSizeAndAnchors(buttonListContentGO, _tasButtonGO, 150f, 30f, "right", -150f - (horizontalOffset * 3));
                        _tasSelectButton = _tasButtonGO.transform.GetComponent<UIDynamicButton>();
                        _tasSelectButton.buttonText.fontSize = 14;
                        _tasSelectButton.label = atomName;

                        _tasSelectButton.button.onClick.AddListener(delegate
                        {
                            try
                            {
                                if (!GridsDisplay._uiTargetAtomSelector.keepOpenJSB.val)
                                {
                                    GameControlUI.gameControlDisplayMode = GameControlDisplayModes.standard;
                                    GridsDisplay._uiTargetAtomSelector.DestroyUI();
                                }
                                
                                GameControlUI.RefreshWristUIButtonGrid();
                                List<Atom> targetAtoms = new List<Atom>();
                                targetAtoms.Add(SuperController.singleton.GetAtomByUid(atomName));
                                buttonCallback.Invoke(targetAtoms);
                            }
                            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
                        });

                    }
                }
                catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
            }
            public void UpdateAtomName(string atomName)
            {
                _tasSelectButton.label = atomName;
            }
            public void DestroyTASItem()
            {
                GameObject.Destroy(_tasButtonGO);
            }
        }

        private GridsCanvas _uiButtonCanvas;
        public GameObject _tasScrollListGO;
        private Dictionary<string, UIDTargetAtomSelectorItem> _targetAtomChoicesDictionary;
        private UIAButton.UserAtomSelectedCallback buttonCallback;

        private List<string> _tasAtomChoiceNames;
        public UIDynamicButton _tasCloseButton;
        public JSONStorableBool keepOpenJSB;
        public UIDynamicToggle keepOpenToggle;
        private bool displayKeepOpen = false;

        private float _verticalStartOffset = 30f;

        private Vector2 _scrollBarPosition;

        public TargetAtomSelectorDisplay(GridsCanvas uiButtonCanvas)
        {
            _uiButtonCanvas = uiButtonCanvas;
            _tasAtomChoiceNames = new List<string>();
            _targetAtomChoicesDictionary = new Dictionary<string, UIDTargetAtomSelectorItem>();
            keepOpenJSB = new JSONStorableBool("keepOpen", false);
        }

        public float GetXYDistanceToScrollBar(Vector3 worldSpacePoint, float zThreshold)
        {
            float distance = 100000f;
            ScrollRect scrollRect = _tasScrollListGO.GetComponent<ScrollRect>();
            Scrollbar sb = scrollRect.verticalScrollbar;
            RectTransform rt = sb.transform.GetComponent<RectTransform>();
            Vector3 localSpacePoint = rt.InverseTransformPoint(worldSpacePoint);

            if (localSpacePoint.z < zThreshold) distance = GeometryUtils.DistancePointToRectangle(new Vector2(localSpacePoint.x, localSpacePoint.y), rt.rect);

            return (distance);
        }

        public float GetScrollBarValue()
        {
            ScrollRect scrollRect = _tasScrollListGO.GetComponent<ScrollRect>();
            Scrollbar sb = scrollRect.verticalScrollbar;
            return (sb.value);
        }
        public void OffsetScrollBarValueByYMovement(float initialSBValue, Vector3 initialWorldSpacePoint, Vector3 currentWorldSpacePoint)
        {
            ScrollRect scrollRect = _tasScrollListGO.GetComponent<ScrollRect>();
            Scrollbar sb = scrollRect.verticalScrollbar;
            RectTransform rt = sb.transform.GetComponent<RectTransform>();
            Vector3 localInitialSpacePoint = rt.InverseTransformPoint(initialWorldSpacePoint);
            Vector3 localCurrentSpacePoint = rt.InverseTransformPoint(currentWorldSpacePoint);

            sb.value = Mathf.Clamp(initialSBValue + ((localCurrentSpacePoint.y - localInitialSpacePoint.y) / rt.rect.height), 0f, 1f);
        }

        public void UpdateTAS()
        {
            Color greyCol = new Color(0.8f, 0.8f, 0.8f);
            _tasCloseButton.buttonColor = greyCol;
            keepOpenToggle.backgroundColor = greyCol;
            foreach (KeyValuePair<string, UIDTargetAtomSelectorItem> kvp in _targetAtomChoicesDictionary)
            {
                kvp.Value._tasSelectButton.buttonColor = greyCol;
            }
        }

        public void UpdateTASGazeTouchDistances()
        {
            int index = 0;

            RectTransform scrollListRT = _tasScrollListGO.GetComponent<RectTransform>();
            float checkRectMinY = scrollListRT.rect.yMin;
            float checkRectMaxY = scrollListRT.rect.yMax;

            GameControlUI.SetNearestGazeAndLeapButtons(_tasCloseButton);
            if (keepOpenToggle!=null) GameControlUI.SetNearestGazeAndLeapButtons(keepOpenToggle);
            foreach (KeyValuePair<string, UIDTargetAtomSelectorItem> kvp in _targetAtomChoicesDictionary)
            {
                RectTransform tasButtonRT = kvp.Value._tasButtonGO.GetComponent<RectTransform>();
                Vector3 positionInWorld = tasButtonRT.parent.TransformPoint(tasButtonRT.localPosition);
                Vector3 positionInScroll = scrollListRT.InverseTransformPoint(positionInWorld);

                float childMinY = positionInScroll.y + tasButtonRT.rect.yMin;
                float childMaxY = positionInScroll.y + tasButtonRT.rect.yMax;

                if (childMaxY >= checkRectMinY && childMinY <= checkRectMaxY)
                {
                    GameControlUI.SetNearestGazeAndLeapButtons(kvp.Value._tasSelectButton);
                }

                index++;
            }

        }
        public void AtomRemovedUpdate(string atomName)
        {
            if (_targetAtomChoicesDictionary.ContainsKey(atomName))
            {
                _tasAtomChoiceNames.Remove(atomName);
                _targetAtomChoicesDictionary[atomName].DestroyTASItem();
                _targetAtomChoicesDictionary.Remove(atomName);
            }
        }

        public void UpdateAtomNames(string oldName, string newName)
        {
            if (_targetAtomChoicesDictionary.ContainsKey(oldName))
            {
                UIDTargetAtomSelectorItem tempTAS = _targetAtomChoicesDictionary[oldName];
                tempTAS.UpdateAtomName(newName);
                _targetAtomChoicesDictionary.Remove(oldName);
                _targetAtomChoicesDictionary.Add(newName, tempTAS);
                _tasAtomChoiceNames.Remove(oldName);
                _tasAtomChoiceNames.Add(newName);
            }

        }

        public void SetTargetChoices(List<string> targetAtomNames, UIAButton.UserAtomSelectedCallback _buttonCallback, bool displayKeepOpen)
        {
            buttonCallback = _buttonCallback;
            //           _vrHand = vrHand;
            //           _leapActivated = leapActivated;
            _tasAtomChoiceNames.Clear();
            _tasAtomChoiceNames = targetAtomNames;
            _scrollBarPosition = new Vector2();
            this.displayKeepOpen = displayKeepOpen;
        }

        public void CreateUI()
        {
            float scrollList1Height = 200f;
            float scrollList1Width = 180f;
            float listXoffset = -(scrollList1Width / 4f);
            float selectGroupHeight = 20f;

            _tasScrollListGO = _uiButtonCanvas.CreateScrollableButtonList(0f, (scrollList1Height / 2f) + _verticalStartOffset + selectGroupHeight, scrollList1Width, scrollList1Height, new Color(0.5f, 0.5f, 0.5f, 0.5f));

            ScrollRect scrollRect = _tasScrollListGO.GetComponent<ScrollRect>();
            scrollRect.content.localPosition = _scrollBarPosition;

            _tasCloseButton = _uiButtonCanvas.CreateUIButtonInCanvas("X", 75f, 35f, 0f, 25f, 25f, 16);

            _tasCloseButton.button.onClick.AddListener(delegate
            {
                try
                {
                    GameControlUI.gameControlDisplayMode = GameControlDisplayModes.standard;
                    DestroyUI();
                    GameControlUI.RefreshWristUIButtonGrid();
                }
                catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
            });

            if (displayKeepOpen)
            {
                keepOpenToggle = _uiButtonCanvas.CreateUIToggleInCanvas("Keep Open", -30f, 35f, 0f, 120f, 25f, 14);
                keepOpenJSB.RegisterToggle(keepOpenToggle.toggle);
            }
            

            foreach (string targetAtomName in _tasAtomChoiceNames)
            {
                UIDTargetAtomSelectorItem tasItem = new UIDTargetAtomSelectorItem(_uiButtonCanvas, _tasScrollListGO, targetAtomName, buttonCallback);
                _targetAtomChoicesDictionary.Add(targetAtomName, tasItem);
            }
        }

        public void DestroyUI()
        {
            if (_tasScrollListGO != null)
            {
                ScrollRect scrollRect = _tasScrollListGO.GetComponent<ScrollRect>();
                _scrollBarPosition = scrollRect.content.localPosition;

                foreach (KeyValuePair<string, UIDTargetAtomSelectorItem> kvp in _targetAtomChoicesDictionary) kvp.Value.DestroyTASItem();
                _targetAtomChoicesDictionary.Clear();
                GameObject.Destroy(_tasScrollListGO);
                _tasScrollListGO = null;
            }
            if (_tasCloseButton != null) GameObject.Destroy(_tasCloseButton.gameObject);
            if (keepOpenToggle != null) GameObject.Destroy(keepOpenToggle.gameObject);
        }
    }

    public static class UITools
    {

        public static class AnchorPresets
        {
            public const int none = 0;
            public const int topLeft = 1;
            public const int topMiddle = 2;
            public const int topRight = 3;
            public const int middleLeft = 4;
            public const int centre = 5;
            public const int middleRight = 6;
            public const int bottomLeft = 7;
            public const int bottomMiddle = 8;
            public const int bottomRight = 9;

            public const int vStretchMiddle = 11;
            public const int hStretchMiddle = 12;
            public const int vStretchRight = 13;
            public const int vStretchLeft = 14;
            public const int hStretchTop = 15;
            public const int hStretchBottom = 16;
            public const int vhStretch = 17;

            public static Vector2 GetAnchorMin(int anchorPreset)
            {
                Vector2 anchorMin = new Vector2();

                if (anchorPreset == topLeft) anchorMin = new Vector2(0f, 1f);
                if (anchorPreset == topMiddle) anchorMin = new Vector2(0.5f, 1f);
                if (anchorPreset == topRight) anchorMin = new Vector2(1f, 1f);
                if (anchorPreset == middleLeft) anchorMin = new Vector2(0f, 0.5f);
                if (anchorPreset == centre) anchorMin = new Vector2(0.5f, 0.5f);
                if (anchorPreset == middleRight) anchorMin = new Vector2(1f, 0.5f);
                if (anchorPreset == bottomLeft) anchorMin = new Vector2(0f, 0f);
                if (anchorPreset == bottomMiddle) anchorMin = new Vector2(0.5f, 0f);
                if (anchorPreset == bottomRight) anchorMin = new Vector2(1f, 0f);

                if (anchorPreset == vStretchMiddle) anchorMin = new Vector2(0.5f, 0f);
                if (anchorPreset == vStretchRight) anchorMin = new Vector2(1f, 0f);
                if (anchorPreset == vStretchLeft) anchorMin = new Vector2(0f, 0f);
                if (anchorPreset == hStretchMiddle) anchorMin = new Vector2(0f, 0.5f);
                if (anchorPreset == hStretchTop) anchorMin = new Vector2(0f, 1f);
                if (anchorPreset == hStretchBottom) anchorMin = new Vector2(0f, 0f);
                if (anchorPreset == vhStretch) anchorMin = new Vector2(0f, 0f);

                return anchorMin;
            }

            public static Vector2 GetAnchorMax(int anchorPreset)
            {
                Vector2 anchorMax = new Vector2();

                if (anchorPreset == topLeft) anchorMax = new Vector2(0f, 1f);
                if (anchorPreset == topMiddle) anchorMax = new Vector2(0.5f, 1f);
                if (anchorPreset == topRight) anchorMax = new Vector2(1f, 1f);
                if (anchorPreset == middleLeft) anchorMax = new Vector2(0f, 0.5f);
                if (anchorPreset == centre) anchorMax = new Vector2(0.5f, 0.5f);
                if (anchorPreset == middleRight) anchorMax = new Vector2(1f, 0.5f);
                if (anchorPreset == bottomLeft) anchorMax = new Vector2(0f, 0f);
                if (anchorPreset == bottomMiddle) anchorMax = new Vector2(0.5f, 0f);
                if (anchorPreset == bottomRight) anchorMax = new Vector2(1f, 0f);

                if (anchorPreset == vStretchMiddle) anchorMax = new Vector2(0.5f, 1f);
                if (anchorPreset == vStretchRight) anchorMax = new Vector2(1f, 1f);
                if (anchorPreset == vStretchLeft) anchorMax = new Vector2(0f, 1f);
                if (anchorPreset == hStretchMiddle) anchorMax = new Vector2(1f, 0.5f);
                if (anchorPreset == hStretchTop) anchorMax = new Vector2(1f, 1f);
                if (anchorPreset == hStretchBottom) anchorMax = new Vector2(1f, 0f);
                if (anchorPreset == vhStretch) anchorMax = new Vector2(1f, 1f);

                return anchorMax;
            }

            public static Vector2 GetPivot(int anchorPreset)
            {
                Vector2 pivot = new Vector2();

                if (anchorPreset == topLeft) pivot = new Vector2(0f, 1f);
                if (anchorPreset == topMiddle) pivot = new Vector2(0.5f, 1f);
                if (anchorPreset == topRight) pivot = new Vector2(1f, 1f);
                if (anchorPreset == middleLeft) pivot = new Vector2(0f, 0.5f);
                if (anchorPreset == centre) pivot = new Vector2(0.5f, 0.5f);
                if (anchorPreset == middleRight) pivot = new Vector2(1f, 0.5f);
                if (anchorPreset == bottomLeft) pivot = new Vector2(0f, 0f);
                if (anchorPreset == bottomMiddle) pivot = new Vector2(0.5f, 0f);
                if (anchorPreset == bottomRight) pivot = new Vector2(1f, 0f);

                if (anchorPreset == vStretchMiddle) pivot = new Vector2(0.5f, 1f);
                if (anchorPreset == hStretchMiddle) pivot = new Vector2(0f, 0.5f);
                if (anchorPreset == vStretchRight) pivot = new Vector2(1f, 1f);
                if (anchorPreset == vStretchLeft) pivot = new Vector2(0f, 1f);
                if (anchorPreset == hStretchTop) pivot = new Vector2(0f, 1f);
                if (anchorPreset == hStretchBottom) pivot = new Vector2(0f, 0f);
                if (anchorPreset == vhStretch) pivot = new Vector2(0f, 1f);

                return pivot;
            }

        }

        public static UIDynamicPopup CreateScrollablePopup(Transform parentTransform, JSONStorableStringChooser jssc, float width, float height, string label = null, int labelFontSize = 20, float anchorXOffset = 0, float anchorYOffset = 0, float popupHeight = 300f, bool popupAbove = false, bool disableSlider = false, int anchorPreset = AnchorPresets.none, float labelProportion = 0.4f, float popupRowHeight = 15f, int popupFontSize = 20)
        {
            Transform transform = GameObject.Instantiate<Transform>(UIAGlobals.mvrScript.manager.configurableScrollablePopupPrefab);

            GameObject popupGO = transform.gameObject;
            UIDynamicPopup uiPopup = popupGO.transform.GetComponent<UIDynamicPopup>();

            jssc.RegisterPopup(uiPopup.popup);
            if (label != null) uiPopup.popup.label = label;

            float sliderHeight = 10f;
            if (disableSlider) sliderHeight = 0f;

            uiPopup.popup.selectColor = Color.cyan;
            transform.SetParent(parentTransform, false);

            ResizeRecTransform(popupGO, width, height, anchorXOffset, anchorYOffset, anchorPreset);

            GameObject dropDownButtonGO = transform.Find("Button").gameObject;
            ReformatUnityButton(dropDownButtonGO, width * (1 - labelProportion), (height * .9f) - sliderHeight, popupFontSize, -2f, 1.5f, AnchorPresets.middleRight, true);

            GameObject textGO = transform.Find("Text").gameObject;
            ReformatText(textGO, width * (labelProportion), height * .9f, labelFontSize, 0f, 0f, AnchorPresets.middleLeft);

            GameObject popupPanelGO = transform.Find("PopupPanel").gameObject;
            GameObject scrollViewGO = popupPanelGO.transform.Find("Scroll View").gameObject;

            ScrollRect scrollRect = scrollViewGO.GetComponent<ScrollRect>();
            scrollRect.verticalScrollbarVisibility = ScrollRect.ScrollbarVisibility.AutoHideAndExpandViewport;

            if (popupAbove) ReformatPopupPanel(popupPanelGO, width * (1 - labelProportion), popupHeight, popupFontSize, 0f, height, anchorPreset: AnchorPresets.bottomRight, popupRowHeight: popupRowHeight);
            else ReformatPopupPanel(popupPanelGO, width * (1 - labelProportion), popupHeight, popupFontSize, 0f, -height, anchorPreset: AnchorPresets.topRight, popupRowHeight: popupRowHeight);

            GameObject sliderGO = transform.Find("Slider").gameObject;
            ReformatScrollPopupSlider(sliderGO, width * (1 - labelProportion), sliderHeight, 0f, 0f, anchorPreset: AnchorPresets.topRight, clearSprite: false);

            return uiPopup;
        }

        private static void ReformatPopupPanel(GameObject popupPanelGO, float width, float height, int fontSize = 12, float anchorXOffset = 0, float anchorYOffset = 0, int anchorPreset = AnchorPresets.none, float popupRowHeight = 15f)
        {
            float scrollBarWidth = 10f;
            ResizeRecTransform(popupPanelGO, width, height, anchorXOffset, anchorYOffset, anchorPreset);

            GameObject buttonTemplateGO = popupPanelGO.transform.Find("PopupValueButtonTemplate").gameObject;
            ReformatUnityButton(buttonTemplateGO, width - scrollBarWidth, popupRowHeight, fontSize, clearSprite: true, anchorPreset: AnchorPresets.none);

            if (popupPanelGO.transform.Find("Scroll View") != null)
            {
                ReformatScrollView(popupPanelGO.transform.Find("Scroll View").gameObject, width, scrollBarWidth, height);
            }
        }

        private static void ReformatScrollView(GameObject scrollViewGO, float width, float vScrollBarWidth, float height = 25, float hScrollBarHeight = 0)
        {
            GameObject viewPortGO = scrollViewGO.transform.Find("Viewport").gameObject;
            ResizeRecTransform(viewPortGO, width - vScrollBarWidth, height - hScrollBarHeight);

            RectTransform viewPortRT = viewPortGO.transform.GetComponent<RectTransform>();
            viewPortRT.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, height - hScrollBarHeight / 2);
            viewPortRT.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, width - vScrollBarWidth / 2);

            if (scrollViewGO.transform.Find("Scrollbar Vertical") != null)
            {
                GameObject scrollBarVerticalGO = scrollViewGO.transform.Find("Scrollbar Vertical").gameObject;
                Scrollbar sb = scrollBarVerticalGO.transform.GetComponent<Scrollbar>();
                sb.enabled = vScrollBarWidth > 0f;

                RectTransform scrollBarRT = scrollBarVerticalGO.transform.GetComponent<RectTransform>();
                scrollBarRT.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, vScrollBarWidth);
                scrollBarRT.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, height);
            }
            if (scrollViewGO.transform.Find("Scrollbar Horizontal") != null)
            {
                GameObject scrollBarVerticalGO = scrollViewGO.transform.Find("Scrollbar Horizontal").gameObject;
                Scrollbar sb = scrollBarVerticalGO.transform.GetComponent<Scrollbar>();
                sb.enabled = hScrollBarHeight > 0f;
                RectTransform scrollBarRT = scrollBarVerticalGO.transform.GetComponent<RectTransform>();
                scrollBarRT.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, hScrollBarHeight);
                scrollBarRT.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, width);

            }
        }
        private static void ReformatScrollPopupSlider(GameObject sliderGO, float width, float height, float anchorXOffset = 0, float anchorYOffset = 0, int anchorPreset = AnchorPresets.none, bool clearSprite = true)
        {
            ResizeRecTransform(sliderGO, width, height, anchorXOffset, anchorYOffset, anchorPreset);

            GameObject handleGO = sliderGO.transform.Find("Handle Slide Area/Handle").gameObject;
            Image handleImage = handleGO.transform.GetComponent<Image>();
            if (clearSprite) handleImage.sprite = null;
        }
        private static void ReformatUnityButton(GameObject buttonGO, float width, float height, int fontSize = 12, float anchorXOffset = 0, float anchorYOffset = 0, int anchorPreset = AnchorPresets.none, bool clearSprite = true, TextAnchor textAlignment = TextAnchor.MiddleCenter, bool resizeTextForBestFit = false)
        {
            ResizeRecTransform(buttonGO, width, height, anchorXOffset, anchorYOffset, anchorPreset);

            if (clearSprite)
            {
                Image buttonImage = buttonGO.transform.GetComponent<Image>();
                buttonImage.sprite = null;
            }

            Transform textTransform = buttonGO.transform.Find("Text");
            ReformatText(textTransform.gameObject, width - 4f, height, fontSize, 2f, 0f, AnchorPresets.centre, textAlignment, resizeTextForBestFit);
        }

        public static GameObject AddChildImage(Transform parentTransform, Color color, Vector2 anchorMin, Vector2 anchorMax, float horizontalSize, float verticalSize, Vector2 anchoredPositionOffset, Vector2 pivot, Sprite sprite = null, bool enabled = true)
        {
            GameObject go = new GameObject();
            go.transform.SetParent(parentTransform, false);

            AddImageComponent(go, color, sprite, enabled);
            RectTransform rt = go.GetComponent<RectTransform>();
            rt.sizeDelta = Vector2.zero;
            rt.anchorMin = anchorMin;
            rt.anchorMax = anchorMax;
            rt.pivot = pivot;
            if (verticalSize != 0f) rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, verticalSize);
            if (horizontalSize != 0f) rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, horizontalSize);

            rt.anchoredPosition = anchoredPositionOffset;

            return go;
        }

        public static GameObject AddChildImage(GameObject parentGO, Color color, Vector2 anchorMin, Vector2 anchorMax, float horizontalSize, float verticalSize, Vector2 anchoredPositionOffset, Vector2 pivot, Sprite sprite = null, bool enabled = true)
        {
            return AddChildImage(parentGO.transform, color, anchorMin, anchorMax, horizontalSize, verticalSize, anchoredPositionOffset, pivot, sprite, enabled);
        }
        public static GameObject AddChildGOImage(GameObject parentGO, Color color, int anchorPreset, float horizontalSize, float verticalSize, Vector2 anchoredPositionOffset, Sprite sprite = null, bool enabled = true)
        {
            return AddChildImage(parentGO.transform, color, AnchorPresets.GetAnchorMin(anchorPreset), AnchorPresets.GetAnchorMax(anchorPreset), horizontalSize, verticalSize, anchoredPositionOffset, AnchorPresets.GetPivot(anchorPreset), sprite, enabled);
        }

        public static GameObject AddChildImage(Transform parentTransform, Color color, int anchorPreset, float horizontalSize, float verticalSize, Vector2 anchoredPositionOffset, Sprite sprite = null, bool enabled = true)
        {
            return AddChildImage(parentTransform, color, AnchorPresets.GetAnchorMin(anchorPreset), AnchorPresets.GetAnchorMax(anchorPreset), horizontalSize, verticalSize, anchoredPositionOffset, AnchorPresets.GetPivot(anchorPreset), sprite, enabled);
        }

        public static GameObject CreateBasicTextBox(Transform parentTransform, string textString, float width, float height, int fontSize = 20, float anchorXOffset = 0, float anchorYOffset = 0, int anchorPreset = AnchorPresets.none, bool resizeTextForBestFit = false, int minFontSize = 20, HorizontalWrapMode horizontalWrapMode = HorizontalWrapMode.Wrap, Color? backgoundColor = null)
        {
            GameObject imageGO = AddChildImage(parentTransform, backgoundColor ?? Color.gray, anchorPreset, width, height, new Vector2(anchorXOffset, anchorYOffset));
            var textGo = new GameObject("text");
            textGo.transform.SetParent(imageGO.transform, false);
            textGo.AddComponent<RectTransform>();
            Text text = textGo.AddComponent<Text>();
            Font selectedFont = (Font)Resources.GetBuiltinResource(typeof(Font), "Arial.ttf");
            text.font = selectedFont;
            ReformatText(textGo, width - 4f, height, fontSize, 2f, 0, AnchorPresets.centre, TextAnchor.MiddleLeft, resizeTextForBestFit);

            text.horizontalOverflow = horizontalWrapMode;
            //           text.resizeTextForBestFit = resizeTextForBestFit;

            if (resizeTextForBestFit)
            {
                text.resizeTextMaxSize = fontSize;
                text.resizeTextMinSize = minFontSize;
            }
            else text.fontSize = fontSize;

            text.text = textString;

            text.color = Color.black;
            return imageGO;
        }

        public static GameObject CreateBasicTextBox(Transform parentTrans, JSONStorableString jss, float width, float height, int fontSize = 20, float anchorXOffset = 0, float anchorYOffset = 0, int anchorPreset = AnchorPresets.none, bool resizeTextForBestFit = false, int minFontSize = 20, HorizontalWrapMode horizontalWrapMode = HorizontalWrapMode.Wrap, Color? backgoundColor = null)
        {
            GameObject textImageGO = CreateBasicTextBox(parentTrans, jss.val, width, height, fontSize, anchorXOffset, anchorYOffset, anchorPreset, resizeTextForBestFit, minFontSize, horizontalWrapMode, backgoundColor);

            GameObject textGO = textImageGO.transform.Find("text").gameObject;

            jss.RegisterText(textGO.GetComponent<Text>());

            return textImageGO;
        }

        public static void ReformatText(GameObject textGO, float width, float height, int fontSize = 20, float anchorXOffset = 0, float anchorYOffset = 0, int anchorPreset = AnchorPresets.none, TextAnchor textAlginment = TextAnchor.MiddleCenter, bool resizeTextForBestFit = false)
        {

            RectTransform textRT = textGO.transform.GetComponent<RectTransform>();
            textRT.localScale = new Vector3(0.5f, 0.5f, 0.5f);
            width = width * 2;
            height = height * 2;

            ResizeRecTransform(textGO, width, height, anchorXOffset, anchorYOffset, anchorPreset);

            Text text = textGO.transform.GetComponent<Text>();
            text.alignment = textAlginment;
            text.fontSize = fontSize;
            text.resizeTextForBestFit = resizeTextForBestFit;
        }

        public static void ResizeRecTransform(GameObject go, float width, float height, float anchorXOffset = 0, float anchorYOffset = 0, int anchorPreset = AnchorPresets.none)
        {
            RectTransform rt = go.GetComponent<RectTransform>();
            ResizeRecTransform(rt, width, height, anchorXOffset, anchorYOffset, anchorPreset);

        }
        public static void ResizeRecTransform(RectTransform rt, float width, float height, float anchorXOffset = 0, float anchorYOffset = 0, int anchorPreset = AnchorPresets.none)
        {
            rt.sizeDelta = Vector2.zero;
            if (anchorPreset != AnchorPresets.none)
            {

                rt.anchorMin = AnchorPresets.GetAnchorMin(anchorPreset);
                rt.anchorMax = AnchorPresets.GetAnchorMax(anchorPreset);
                rt.pivot = AnchorPresets.GetPivot(anchorPreset);
            }
            rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, width);
            rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, height);
            rt.anchoredPosition = new Vector2(anchorXOffset, anchorYOffset);
        }

        public static GameObject CreateToggle(Transform parent, float xAnchorOffset, float yAnchorOffset, float width, float height, string label = "", int fontSize = 12)
        {

            Transform transform = GameObject.Instantiate<Transform>(UIAGlobals.mvrScript.manager.configurableTogglePrefab);

            GameObject toggleGO = transform.gameObject;
            transform.SetParent(parent, false);
            transform.gameObject.SetActive(true);

            RectTransform rt = transform.GetComponent<RectTransform>();
            rt.pivot = new Vector2(0f, 0f);
            rt.localPosition = new Vector2(xAnchorOffset, yAnchorOffset);

            rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, width);
            rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, height);

            UIDynamicToggle uiToggle = toggleGO.transform.GetComponent<UIDynamicToggle>();
            uiToggle.label = label;
            uiToggle.labelText.fontSize = fontSize;

            GameObject checkBoxGO = transform.Find("Background").gameObject;
            RectTransform checkBoxRT = checkBoxGO.transform.GetComponent<RectTransform>();
            checkBoxRT.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, height);
            checkBoxRT.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, height);
            checkBoxRT.anchoredPosition = new Vector2(height / 2f, 0f);

            GameObject checkMarkGO = transform.Find("Background/Checkmark").gameObject;
            RectTransform checkMarkRT = checkMarkGO.transform.GetComponent<RectTransform>();
            checkMarkRT.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, height);
            checkMarkRT.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, height);

            GameObject toggleLabelGO = transform.Find("Label").gameObject;
            RectTransform toggleLabelRT = toggleLabelGO.transform.GetComponent<RectTransform>();
            toggleLabelRT.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, width - height - 5f);
            toggleLabelRT.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, height - 2f);
            toggleLabelRT.pivot = new Vector2(0f, 0f);
            toggleLabelRT.localPosition = new Vector2(height + 2f, 1f);

            return (toggleGO);
        }

        private static void AddImageComponent(GameObject go, Color color, Sprite sprite = null, bool enabled = true)
        {
            UnityEngine.UI.Image image = go.AddComponent<UnityEngine.UI.Image>();
            image.color = color;
            image.sprite = sprite;
            image.type = UnityEngine.UI.Image.Type.Sliced;
            image.enabled = enabled;
        }

        public static GameObject CreateScrollableButtonList(Transform parent, float xAnchorOffset, float yAnchorOffset, float width, float height, Color backgroundColor, float scrollbarWidth=10f)
        {
            GameObject scrollableButtonListGO = new GameObject();

            AddImageComponent(scrollableButtonListGO, backgroundColor);
            RectTransform imageRT = scrollableButtonListGO.GetComponent<RectTransform>();
            imageRT.anchoredPosition = new Vector2(xAnchorOffset, yAnchorOffset);
            imageRT.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, width);
            imageRT.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, height);

            GameObject buttonListViewportGO = CreateButtonListViewport(scrollableButtonListGO, width - 20f, height - 10f, 5f, 5f);
            buttonListViewportGO.name = "buttonListViewportGO";

            GameObject buttonListScrollbarGO = CreateScrollBar(scrollableButtonListGO, scrollbarWidth, height - 10f, -(scrollbarWidth/4f), 5f, Scrollbar.Direction.BottomToTop);

            ScrollRect scrollRect = scrollableButtonListGO.AddComponent<ScrollRect>();
            GameObject buttonListContentGO = buttonListViewportGO.transform.Find("buttonListContentGO").gameObject;
            scrollRect.content = buttonListContentGO.GetComponent<RectTransform>();
            scrollRect.horizontal = false;
            scrollRect.vertical = true;
            scrollRect.movementType = ScrollRect.MovementType.Clamped;
            scrollRect.viewport = buttonListViewportGO.GetComponent<RectTransform>();
            scrollRect.verticalScrollbar = buttonListScrollbarGO.GetComponent<Scrollbar>();
            scrollRect.verticalScrollbarVisibility = ScrollRect.ScrollbarVisibility.Permanent;

            scrollableButtonListGO.transform.SetParent(parent, false);
            scrollableButtonListGO.SetActive(true);

            return (scrollableButtonListGO);
        }
        private static GameObject CreateScrollBar(GameObject parentGO, float width, float height, float xAnchorOffset, float yAnchorOffset, Scrollbar.Direction scrollbarDirection)
        {
            Sprite backgroundSprite = Resources.FindObjectsOfTypeAll<Sprite>().First(sprite => sprite.name == "Background");
            Sprite uiSprite = Resources.FindObjectsOfTypeAll<Sprite>().First(sprite => sprite.name == "UISprite");

            GameObject scrollbarGO = new GameObject();
            Scrollbar scrollbar = scrollbarGO.AddComponent<Scrollbar>();
            AddImageComponent(scrollbarGO, new Color(1f, 1f, 1f), backgroundSprite);
            scrollbar.SetDirection(scrollbarDirection, false);
            scrollbarGO.transform.SetParent(parentGO.transform, false);
            scrollbarGO.SetActive(true);
            SetSizeAndAnchors(parentGO, scrollbarGO, width, height, "topright", xAnchorOffset, yAnchorOffset);

            GameObject scrollbarSlidingAreaGO = new GameObject();
            scrollbarSlidingAreaGO.AddComponent<RectTransform>();
            scrollbarSlidingAreaGO.transform.SetParent(scrollbarGO.transform, false);
            scrollbarSlidingAreaGO.SetActive(true);
            SetSizeAndAnchors(scrollbarGO, scrollbarSlidingAreaGO, width, height, "center", 0, 0);

            GameObject scrollbarHandleGO = new GameObject();
            scrollbarHandleGO.transform.SetParent(scrollbarSlidingAreaGO.transform, false);
            AddImageComponent(scrollbarHandleGO, new Color(1f, 1f, 1f), uiSprite);
            SetSizeAndAnchors(scrollbarSlidingAreaGO, scrollbarHandleGO, 0f, 0f, "center", 0, 0);

            scrollbar.targetGraphic = scrollbarHandleGO.GetComponent<UnityEngine.UI.Image>();
            scrollbar.handleRect = scrollbarHandleGO.GetComponent<RectTransform>(); ;

            return (scrollbarGO);
        }

        public static void SetSizeAndAnchors(GameObject parent, GameObject child, float width, float height, string placement = "center", float offsetX = 0, float offsetY = 0, float pivotX = 0.5f, float pivotY = 0.5f)
        {
            RectTransform childRect = child.GetComponent<RectTransform>();
            childRect.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, width);
            childRect.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, height);
            SetAnchors(parent, child, placement, offsetX, offsetY, pivotX, pivotY);

        }
        private static void SetAnchors(RectTransform parent, RectTransform child, string placement = "center", float offsetX = 0, float offsetY = 0, float pivotX = 0.5f, float pivotY = 0.5f)
        {
            if (parent == null)
            {
                parent = child.parent.GetComponent<RectTransform>();
            }
            Vector2 pivotPoint = new Vector2(pivotX, pivotY);
            child.pivot = pivotPoint;

            switch (placement)
            {
                case "center":
                    child.transform.localPosition = new Vector3(0, 0, 0);
                    break;
                case "left":
                    child.transform.localPosition = new Vector3(parent.rect.xMin - child.rect.xMin + offsetX, 0 + offsetY);
                    break;
                case "right":
                    child.transform.localPosition = new Vector3(parent.rect.xMax - child.rect.xMax + offsetX, 0 - offsetY);
                    break;
                case "top":
                    child.transform.localPosition = new Vector3(0, parent.rect.yMax - child.rect.yMax + offsetX, 0 - offsetY);
                    break;
                case "bottom":
                    child.transform.localPosition = new Vector3(0, parent.rect.yMin - child.rect.yMin + offsetX, 0 + offsetY);
                    break;

                case "topleft":
                    child.transform.localPosition = new Vector3(parent.rect.xMin - child.rect.xMin + offsetX, parent.rect.yMax - child.rect.yMax - offsetY);
                    break;
                case "topright":
                    child.transform.localPosition = new Vector3(parent.rect.xMax - child.rect.xMax + offsetX, parent.rect.yMax - child.rect.yMax - offsetY);
                    break;

                case "bottomleft":
                    child.transform.localPosition = new Vector3(parent.rect.xMin - child.rect.xMin + offsetX, parent.rect.yMin - child.rect.yMin + offsetY);
                    break;
                case "bottomright":
                    child.transform.localPosition = new Vector3(parent.rect.xMax - child.rect.xMax + offsetX, parent.rect.yMin - child.rect.yMin + offsetY);
                    break;
                //Outside Edges
                case "leftout":
                    child.transform.localPosition = new Vector3(parent.rect.xMin + child.rect.xMin + offsetX, 0 + offsetY);
                    break;
                case "rightout":
                    child.transform.localPosition = new Vector3(parent.rect.xMax + child.rect.xMax + offsetX, 0 - offsetY);
                    break;
                case "topout":
                    child.transform.localPosition = new Vector3(0, parent.rect.yMax + child.rect.yMax + offsetX, 0 - offsetY);
                    break;
                case "bottomout":
                    child.transform.localPosition = new Vector3(0, parent.rect.yMin + child.rect.yMin + offsetX, 0 + offsetY);
                    break;
            }

        }
        public static void SetAnchors(GameObject parent, GameObject child, string placement = "center", float offsetX = 0, float offsetY = 0, float pivotX = 0.5f, float pivotY = 0.5f)
        {
            RectTransform parentRect = null;
            if (parent != null) parentRect = parent.GetComponent<RectTransform>();
            RectTransform childRect = child.GetComponent<RectTransform>();
            SetAnchors(parentRect, childRect, placement, offsetX, offsetY, pivotX, pivotY);
        }

        public static Sprite defaultButtonSprite = null;
        public static GameObject CreateUIButtonInParent(Transform parentTransform)
        {
            Transform buttonTransform = UnityEngine.Object.Instantiate(UIAGlobals.mvrScript.manager.configurableButtonPrefab.transform, parentTransform, false);
            if (defaultButtonSprite==null) defaultButtonSprite = buttonTransform.GetComponent<UIDynamicButton>().buttonImage.sprite;
            GameObject buttonGO = buttonTransform.gameObject;
            buttonTransform.SetParent(parentTransform, false);
            return (buttonGO);

        }
        public static GameObject CreateUIContainerImageInParentGO(Transform parentTransform, float width, float height, Color panelColor, string anchorAlignment= "center")
        {
            GameObject containerImageGO = new GameObject();

            AddImageComponent(containerImageGO, panelColor);

            containerImageGO.SetActive(true);

            SetSizeAndAnchors(parentTransform.gameObject, containerImageGO, width, height, "center", 0f, 0f);

            containerImageGO.transform.SetParent(parentTransform, false);
            return (containerImageGO);
        }

        private static GameObject CreateButtonListViewport(GameObject parentGO, float width, float height, float xAnchorOffset, float yAnchorOffset)
        {
            GameObject buttonListViewportGO = new GameObject();


            AddImageComponent(buttonListViewportGO, new Color(1f, 1f, 1f));
            buttonListViewportGO.transform.SetParent(parentGO.transform, false);
            buttonListViewportGO.SetActive(true);
            Mask blvpMask = buttonListViewportGO.AddComponent<Mask>();
            blvpMask.showMaskGraphic = false;
            SetSizeAndAnchors(parentGO, buttonListViewportGO, width, height, "topleft", xAnchorOffset, yAnchorOffset);

            GameObject buttonListContentGO = new GameObject();
            AddImageComponent(buttonListContentGO, new Color(0f, 0f, 1f), null, false);
            buttonListContentGO.transform.SetParent(buttonListViewportGO.transform, false);
            buttonListContentGO.SetActive(true);

            SetSizeAndAnchors(buttonListViewportGO, buttonListContentGO, 0.5f, height / 4, "center", xAnchorOffset, yAnchorOffset, 0.5f, 1f);

            buttonListContentGO.name = "buttonListContentGO";

            VerticalLayoutGroup vlg = buttonListContentGO.AddComponent<VerticalLayoutGroup>();
            vlg.childAlignment = TextAnchor.UpperCenter;
            vlg.childForceExpandHeight = true;
            vlg.childForceExpandWidth = true;
            vlg.childControlHeight = false;
            vlg.childControlWidth = false;
            vlg.spacing = 0f;

            ContentSizeFitter csf = buttonListContentGO.AddComponent<ContentSizeFitter>();
            csf.horizontalFit = ContentSizeFitter.FitMode.MinSize;
            csf.verticalFit = ContentSizeFitter.FitMode.MinSize;

            return (buttonListViewportGO);
        }
    }


    public class CanvasBase
    {
        public CanvasBase()
        {
            GameObject canvasObject = new GameObject();
            _canvas = canvasObject.AddComponent<Canvas>();
            _canvas.pixelPerfect = false;
            _canvas.renderMode = RenderMode.WorldSpace;

            SuperController.singleton.AddCanvas(_canvas);

            CanvasScaler cs = canvasObject.AddComponent<CanvasScaler>();
            if (UIAGlobals.vrActive) cs.dynamicPixelsPerUnit = 4f;
            else cs.dynamicPixelsPerUnit = 1.3f;

            _canvasGroup = canvasObject.AddComponent<CanvasGroup>();

            cs.scaleFactor = 80f;

            _canvas.sortingOrder = -3;

            AdjustCanvasForWorldScale();

            GraphicRaycaster rayCaster = canvasObject.AddComponent<GraphicRaycaster>();
        }
        public GameObject CreateScrollableButtonList(float xAnchorOffset, float yAnchorOffset, float width, float height, Color backgroundColor, float scrollbarWidth =10f)
        {
            return (UITools.CreateScrollableButtonList(_canvas.transform,xAnchorOffset,yAnchorOffset,width,height,backgroundColor, scrollbarWidth));
        }

        private bool _attachedVAMUI = false;
        public bool attachedVAMUI
        {
            get { return _attachedVAMUI; }
            set
            {
                if (_attachedVAMUI != value) GridsControlDisplay.ResetButtonPositions();
                _attachedVAMUI = value;
                AdjustCanvasForWorldScale();

            }
        }

        public bool isVisible { get { return _canvas.enabled; } set { _canvas.enabled = value; } }

        public void OnDestroy()
        {
            if (SuperController.singleton != null) SuperController.singleton.RemoveCanvas(_canvas);
            _canvas.transform.SetParent(null, false);

            if (_canvas.gameObject != null) GameObject.Destroy(_canvas.gameObject);
        }

        public UIDynamicButton CreateUIButtonInCanvas(string label, float x, float y, float z = 0, float w = 0, float h = 0, int fontSize = 16)
        {
            Transform transform = UnityEngine.Object.Instantiate(UIAGlobals.mvrScript.manager.configurableButtonPrefab.transform);

            transform.SetParent(_canvas.transform, false);
            transform.gameObject.SetActive(true);

            UIDynamicButton uiButton = transform.GetComponent<UIDynamicButton>();
            uiButton.label = label;
            uiButton.buttonText.fontSize = fontSize;

            if (w > 0) uiButton.button.GetComponent<RectTransform>().SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, w);
            if (h > 0) uiButton.button.GetComponent<RectTransform>().SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, h);

            transform.localPosition = new Vector3(x, y, z);

            return uiButton;
        }

        public UIDynamicToggle CreateUIToggleInCanvas(string label, float x, float y, float z = 0, float w = 0, float h = 0, int fontSize = 16)
        {
            Transform transform = UnityEngine.Object.Instantiate(UIAGlobals.mvrScript.manager.configurableTogglePrefab.transform);

            transform.SetParent(_canvas.transform, false);
            transform.gameObject.SetActive(true);

            UIDynamicToggle uiToggle = transform.GetComponent<UIDynamicToggle>();
            uiToggle.label = label;
            uiToggle.labelText.fontSize = fontSize;

            GameObject checkBoxGO = transform.Find("Background").gameObject;
            GameObject checkMarkGO = transform.Find("Background/Checkmark").gameObject;
            GameObject labelTextGO = transform.Find("Label").gameObject;

            if (w > 0) uiToggle.toggle.GetComponent<RectTransform>().SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, w);
            if (h > 0)
            {
                uiToggle.toggle.GetComponent<RectTransform>().SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, h);
                checkBoxGO.GetComponent<RectTransform>().SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, h);
                checkBoxGO.GetComponent<RectTransform>().SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, h);
                checkMarkGO.GetComponent<RectTransform>().SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, h);
                checkMarkGO.GetComponent<RectTransform>().SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, h);
                labelTextGO.GetComponent<RectTransform>().SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, h);
                labelTextGO.GetComponent<RectTransform>().SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, w-h);
            }

            checkBoxGO.transform.Translate(-0.01f, 0f, 0f);
            labelTextGO.transform.Translate(-0.01f, 0f, 0f);
            transform.localPosition = new Vector3(x, y, z);
            uiToggle.labelText.alignment = TextAnchor.MiddleLeft;

            return uiToggle;
        }


        public UIDynamicTextField CreateTextFieldInCanvas(float x, float y, float z, float width, float height, float margin = 5f)
        {
            Transform textField = GameObject.Instantiate<Transform>(UIAGlobals.mvrScript.manager.configurableTextFieldPrefab);


            textField.SetParent(_canvas.transform, false);
            RectTransform textFieldRectTransform = textField.GetComponent<RectTransform>();

            textFieldRectTransform.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, height);
            textFieldRectTransform.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, width);

            Vector2 pivot = new Vector2();
            pivot.x = 0.5f;
            pivot.y = 0f;

            textFieldRectTransform.pivot = pivot;
            textFieldRectTransform.localPosition = new Vector3(x, y, z);

            UIDynamicTextField uiTextField = textField.GetComponent<UIDynamicTextField>();

            RectTransform rectT = uiTextField.UItext.GetComponent<RectTransform>();
            //                        rectT.anchoredPosition = new Vector2(5f, -4f);
            //            rectT.pivot = new Vector2(0f, 0.5f);
            //            rectT.localPosition = new Vector2(width / 2f, height / 2f);
            rectT.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Left, margin, width - (margin * 2));
            rectT.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Top, margin, height - (margin * 2));
            //            rectT.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, width * 0.8f);
            //            rectT.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, height * 0.8f);        

            return uiTextField;
        }

        public virtual void AdjustCanvasForWorldScale()
        {
            float attachedAdjustFactor = 1f;
            if (UIAGlobals.vrActive && attachedVAMUI) attachedAdjustFactor = 1.7f;

//            _canvas.transform.localPosition = new Vector3(0.26f * worldScale, -0.14f * worldScale, -1f * worldScale);
            _canvas.transform.localScale = new Vector3(UIAConsts._localCanvasScale * worldScale * attachedAdjustFactor, UIAConsts._localCanvasScale * worldScale* attachedAdjustFactor, UIAConsts._localCanvasScale * worldScale * attachedAdjustFactor);
        }


        public virtual void Update()
        {
            AdjustCanvasForWorldScale();
            Transform cameraT = SuperController.singleton.lookCamera.transform;
            Vector3 facingPos;

            if (GameControlUI.gameControlUIMode== GameControlUIModes.vrVAMUI)
            {
                attachedVAMUI = true;
                bool mainHUDActive = SuperController.singleton.mainHUD.gameObject.activeSelf;
                if (mainHUDActive)
                {
                    if (SuperController.singleton.activeUI == SuperController.ActiveUI.None)
                    {
                        
                        isVisible = true;
                        Transform mainHUDTransform = SuperController.singleton.mainHUD.transform;

                        Vector3 mainHUDRotationEuler = SuperController.singleton.mainHUD.transform.rotation.eulerAngles;
                        _canvas.transform.rotation = Quaternion.Euler(mainHUDRotationEuler.x, mainHUDRotationEuler.y, mainHUDRotationEuler.z);
                        _canvas.transform.Rotate(0f, -180f, 0f);

                        _canvas.transform.position = mainHUDTransform.position + (mainHUDTransform.up * 0f * SuperController.singleton.worldScale) + (mainHUDTransform.forward * -0f);
                    }
                    else
                    {
                        if (isVisible) GridsControlDisplay.CloseGrids();
                        isVisible = false;
                        
                    }

                }
                else if (attachedVAMUI)
                {
                    isVisible = false;

                }
            }
            else
            {
                attachedVAMUI = false;
                switch (GameControlUI.gameControlUIMode)
                {
                    case GameControlUIModes.desktopHUD:
                    case GameControlUIModes.vrHUD:
                        if (SuperController.singleton.lookCamera != null)
                        {
                            facingPos = cameraT.position + cameraT.forward * 10000000.0f;
                            if (Camera.main != null)
                            {
                                Vector3 endPos = Camera.main.ViewportToWorldPoint(new Vector3(horizontalHUDOffset, verticalHUDOffset, depthHUDOffset));
                                _canvas.transform.LookAt(facingPos, cameraT.up);
                                _canvas.transform.position = endPos;

                            }
                        }
                        break;
                    case GameControlUIModes.vrStatic:

                        _canvas.transform.position = vrPosition;
                        facingPos = cameraT.position;
                        Vector3 canvasCenter = _canvas.transform.position;

                        if (GameControlUI.gridsActivated)
                        {
                            Vector3 canvasCenterOffset = new Vector3(GridsDisplay._uiMainButtonGridDisplay.GetLeftEdgeXPosition() + (GridsDisplay._uiMainButtonGridDisplay.GetWidth() / 2), (GridsDisplay._uiMainButtonGridDisplay.GetTopEdgeYPosition() - (GridsDisplay._uiMainButtonGridDisplay.GetHeight() / 2)) * UIAConsts._localCanvasScale * worldScale, 0);
                            canvasCenter = canvasCenter + canvasCenterOffset;

                        }
                        if (Vector3.Distance(facingPos, canvasCenter) > (0.25f * worldScale))
                        {
                            Quaternion rotation = Quaternion.LookRotation(canvasCenter - facingPos);
                            Vector3 rotationV = rotation.eulerAngles;
                            rotationV.z = 0f;
                            _canvas.transform.rotation = Quaternion.Euler(rotationV);
                        }
                        break;
                    case GameControlUIModes.vrWrist:
                        _canvas.transform.position = vrPosition;
                        _canvas.transform.rotation = vrRotation;

                        if (VRSettings.hideWristUIWithShowVRHandsJSB.val)
                        {
                            if (VRSettings.vrHandControlJSEnum.val == VRHandControl.rightHand) _canvas.enabled = SuperController.singleton.commonHandModelControl.rightHandEnabled;
                            else if (VRSettings.vrHandControlJSEnum.val == VRHandControl.leftHand) _canvas.enabled = SuperController.singleton.commonHandModelControl.leftHandEnabled;
                        }
                        break;
                }

            }

        }

        public Vector3 vrPosition
        {
            get
            {
                float inverseOffetMultiplier = 1f;
                if (VRSettings.vrHandControlJSEnum.val == VRHandControl.rightHand) inverseOffetMultiplier = -1f;
                Vector3 offsetPosition = (_canvas.transform.up * 0.02f * worldScale) + (_canvas.transform.forward * -0.045f * worldScale) + (_canvas.transform.right * (-0.16f * inverseOffetMultiplier * worldScale));

                if (GameControlUI.gameControlUIMode == GameControlUIModes.vrStatic) return GridsControlDisplay._fixedPinPosition + offsetPosition;
                if (VRSettings.vrHandControlJSEnum.val == VRHandControl.rightHand) return SuperController.singleton.rightHand.position + offsetPosition;
                return SuperController.singleton.leftHand.position + offsetPosition;
            }
        }
        public Quaternion vrRotation
        {
            get
            {
                float inverseMultiplier = 0f;
                if (VRSettings.vrHandControlJSEnum.val == VRHandControl.rightHand) inverseMultiplier = -1f;
                Quaternion offsetRotation = Quaternion.Euler(180 * inverseMultiplier, 90, 188 + (196 * inverseMultiplier));

                if (VRSettings.vrHandControlJSEnum.val == VRHandControl.rightHand) return SuperController.singleton.rightHand.rotation * offsetRotation;
                return SuperController.singleton.leftHand.rotation * offsetRotation;
            }
        }

        public float worldScale {
            get
            {
                if (!UIAGlobals.vrActive) return GameControlSettings.desktopUIScaleJSF.val;
                else return SuperController.singleton.worldScale;
            }
        }

        protected virtual float horizontalHUDOffset
        {
            get
            {
                if (GameControlUI.gameControlUIMode == GameControlUIModes.vrHUD)
                {
                    if (VRSettings.vrPinModeJSEnum.val == VRPinMode.hudRMode)return 0.57f;
                    else if (VRSettings.vrPinModeJSEnum.val == VRPinMode.hudLMode) return 0.43f;
                    else if (VRSettings.vrPinModeJSEnum.val == VRPinMode.hudCustomMode) return VRSettings.vrHUDHorizontalOffsetJSF.val;
                }
                return GameControlSettings.desktopHOffsetJSF.val;
            }
        }
        protected virtual float verticalHUDOffset
        {
            get
            {
                if (GameControlUI.gameControlUIMode == GameControlUIModes.vrHUD)
                {
                    if (VRSettings.vrPinModeJSEnum.val == VRPinMode.hudRMode) return 0.3f;
                    else if (VRSettings.vrPinModeJSEnum.val == VRPinMode.hudLMode) return 0.3f;
                    else if (VRSettings.vrPinModeJSEnum.val == VRPinMode.hudCustomMode) return VRSettings.vrHUDVerticalOffsetJSF.val;
                }
                return GameControlSettings.desktopVOffsetJSF.val;
            }
        }
        public virtual float depthHUDOffset
        {
            get
            {
                if (GameControlUI.gameControlUIMode == GameControlUIModes.desktopHUD) return 1.5f;
                else if (GameControlUI.gameControlUIMode == GameControlUIModes.vrHUD)
                {
                    if (VRSettings.vrPinModeJSEnum.val == VRPinMode.hudCustomMode) return VRSettings.vrHUDDepthOffsetJSF.val;
                }                
                return 0.6f;
            }
        }


        public Canvas _canvas;
        public CanvasGroup _canvasGroup;
    }

    public class GridsCanvas :CanvasBase
    {
        protected override float horizontalHUDOffset
        {
            get
            {
                if (GameControlUI.gameControlUIMode == GameControlUIModes.desktopHUD)
                {                   
                    float width = Mathf.Max(
                        Camera.main.pixelWidth*(1- GameControlSettings.desktopHOffsetJSF.val),
                        GridsDisplay._uiMainButtonGridDisplay.GetWidth()*GameControlSettings.desktopUIScaleJSF.val /1920f * Camera.main.pixelWidth / 2f,
                        GridsDisplay._uiQLBButtonGridDisplay.GetWidth() * GameControlSettings.desktopUIScaleJSF.val / 1920f * Camera.main.pixelWidth / 2f);
                    return (Camera.main.pixelWidth - width) / Camera.main.pixelWidth;
                }
                return base.horizontalHUDOffset;
            }
        }

        public override void Update()
        {
            base.Update();
            switch (GameControlUI.gameControlUIMode)
            {
                case GameControlUIModes.vrWrist:
                    _canvas.transform.Translate(0f, +0.05f * worldScale, 0f, Space.Self);
                    _canvas.transform.Rotate(-50, 0, 0);
                    _canvas.transform.position = _canvas.transform.position + (_canvas.transform.up * 0.02f * worldScale);
                    break;

                case GameControlUIModes.vrVAMUI:
                    _canvas.transform.Translate(0f, -0.02f * worldScale, 0f, Space.Self);
                    break;
            }
        }
    }

    public class VAMUILaunchCanvas : CanvasBase
    {
        public VAMUILaunchCanvas() : base()
        {
            _canvas.gameObject.layer = SuperController.singleton.mainHUD.gameObject.layer;
            _canvas.transform.SetParent(SuperController.singleton.mainHUD, false);

            _canvas.transform.localEulerAngles = new Vector3(32, 180, 0);
        }


        public override void AdjustCanvasForWorldScale()
        {

            _canvas.transform.localScale = new Vector3(UIAConsts._localCanvasScale, UIAConsts._localCanvasScale, UIAConsts._localCanvasScale);
        }

        public override void Update()
        {
            AdjustCanvasForWorldScale();
        }
    }

}
