﻿using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.Collections.Specialized;


namespace JayJayWon
{
    public class JSONStorableInt : JSONStorableFloat
    {
        public delegate void SetValueCallback(int callbackValue);
        public delegate void SetJSIntCallback(JSONStorableInt jsInt);

        public SetValueCallback setValueCallbackFunction;
        public SetJSIntCallback setJSIntCallbackFunction;

        public JSONStorableInt(string paramName, int startingValue, int minValue, int maxValue,  bool constrain = true, bool interactable = true):base(paramName, startingValue, minValue, maxValue, constrain, interactable)
        {
            setCallbackFunction = floatValCallback;
        }

        public JSONStorableInt(string paramName, int startingValue, SetValueCallback callback, int minValue, int maxValue, bool constrain = true, bool interactable = true) : base(paramName, startingValue, minValue, maxValue, constrain, interactable)
        {
            setCallbackFunction = floatValCallback;
            setValueCallbackFunction = callback;
        }

        public JSONStorableInt(string paramName, int startingValue, SetJSIntCallback callback, int minValue, int maxValue, bool constrain = true, bool interactable = true) : base(paramName, startingValue, minValue, maxValue, constrain, interactable)
        {
            setCallbackFunction = floatValCallback;
            setJSIntCallbackFunction = callback;
        }

        private void floatValCallback (float callbackVal)
        {
            base.valNoCallback = Mathf.Round(callbackVal);
            setValueCallbackFunction?.Invoke(val);
            setJSIntCallbackFunction?.Invoke(this);
        }

        public new int val
        {
            get { return (int) base.val; }
            set { base.val = value; }
        }
        public new int valNoCallback
        {
            get { return (int)base.valNoCallback; }
            set { base.valNoCallback = value; }
        }
        public new int defaultVal
        {
            get { return (int)base.defaultVal; }
            set { base.defaultVal = value;}
        }
        public new int max
        {
            get { return (int)base.max; }
            set { base.max = value; }
        }
        public new int min
        {
            get { return (int)base.min; }
            set { base.min = value; }
        }
    }

    public class JSONStorableMultiEnumStringChooser : JSONStorableStringChooser
    {
        public delegate void SetEnumCallback(int enumValue);
        public delegate void SetMainValCallback(string mainValue);

        public SetEnumCallback setTopEnumCallbackFunction;
        public SetEnumCallback setBottomEnumCallbackFunction;
        public SetMainValCallback setMainValCallbackFunction;

        public const int topEnumValFlag = 1;
        public const int mainValFlag = 2;
        public const int bottomEnumValFlag = 3;
        public const int unrecognisedValFlag = 4;

        protected const string topPrefix = "T";
        protected const string mainPrefix = "M";
        protected const string bottomPrefix = "B";

        protected SortedDictionary<int, string> topEnumChoicesDict;
        protected Dictionary<int, string> topEnumStoreValsDict;
        protected Dictionary<string, int> topStoreValsEnumDict;

        protected List<string> mainChoices;
        protected List<string> mainDisplayChoices;
        public string displayChoicePreFix { get; protected set; }
        public string displayChoicePostFix { get; protected set; }

        protected SortedDictionary<int, string> bottomEnumChoicesDict;
        protected Dictionary<int, string> bottomEnumStoreValsDict;
        protected Dictionary<string, int> bottomStoreValsEnumDict;

        public JSONStorableMultiEnumStringChooser(string paramName, List<string> _choices, List<string> _displayChoices =null, string _displayChoicePreFix=null, string _displayChoicePostFix=null, string topEnumManifestName=null, string bottomEnumManifestName=null) : base(paramName, null, null, "")
        {
            Init(_choices, _displayChoices, _displayChoicePreFix, _displayChoicePostFix, topEnumManifestName, bottomEnumManifestName);
        }
        public JSONStorableMultiEnumStringChooser(string paramName, List<string> _choices, List<string> _displayChoices = null, string _displayChoicePreFix = null, string _displayChoicePostFix = null, SetMainValCallback mainValCallbackFunction=null,string topEnumManifestName = null, SetEnumCallback topEnumCallbackFunction=null, string bottomEnumManifestName = null, SetEnumCallback bottomEnumCallbackFunction = null) : base(paramName, null, null, "")
        {
            Init(_choices, _displayChoices, _displayChoicePreFix, _displayChoicePostFix, topEnumManifestName, bottomEnumManifestName);
            setMainValCallbackFunction = mainValCallbackFunction;
            setTopEnumCallbackFunction = topEnumCallbackFunction;
            setBottomEnumCallbackFunction = bottomEnumCallbackFunction;

            setCallbackFunction = ValChangeCallback;
        }

        protected void ValChangeCallback(string baseVal)
        {

            if (baseVal.StartsWith(mainPrefix)) setMainValCallbackFunction?.Invoke(mainVal);
            else if (baseVal.StartsWith(topPrefix)) setTopEnumCallbackFunction?.Invoke(valTopEnum);
            else if (baseVal.StartsWith(bottomPrefix)) setBottomEnumCallbackFunction?.Invoke(valBottomEnum);
        }
        private void Init(List<string> _choices, List<string> _displayChoices , string _displayChoicePreFix, string _displayChoicePostFix, string _topEnumManifestName, string _bottomEnumManifestName)
        {
            topEnumChoicesDict = new SortedDictionary<int, string>();
            topEnumStoreValsDict = new Dictionary<int, string>();
            topStoreValsEnumDict = new Dictionary<string, int>();
            mainChoices = new List<string>();
            mainDisplayChoices = new List<string>();
            bottomEnumChoicesDict = new SortedDictionary<int, string>();
            bottomEnumStoreValsDict = new Dictionary<int, string>();
            bottomStoreValsEnumDict = new Dictionary<string, int>();

            if (_displayChoices == null) _displayChoices = _choices;
            if (choices!=null && _displayChoices.Count != _choices.Count)
            {
                SuperController.LogError("UIAssist.JSONStorableMultiEnumStringChooser.Con1: displayChoices and choices lists have different counts");
                return;
            }

            if (_choices != null) mainChoices = _choices;
            if (_displayChoices != null) mainDisplayChoices = _displayChoices;
            displayChoicePreFix = _displayChoicePreFix;
            displayChoicePostFix = _displayChoicePostFix;

            if (_topEnumManifestName != null)
            {
                topEnumChoicesDict = EnumManifest.GetEnumsDisplayValDict(_topEnumManifestName);
                topEnumStoreValsDict = EnumManifest.GetEnumsStoreValDict(_topEnumManifestName);
                topStoreValsEnumDict = EnumManifest.GetStoreValEnumsDict(_topEnumManifestName);
            }

            if (_bottomEnumManifestName != null)
            {
                bottomEnumChoicesDict = EnumManifest.GetEnumsDisplayValDict(_bottomEnumManifestName);
                bottomEnumStoreValsDict = EnumManifest.GetEnumsStoreValDict(_bottomEnumManifestName);
                bottomStoreValsEnumDict = EnumManifest.GetStoreValEnumsDict(_bottomEnumManifestName);
            }

            ResetChoices();
            if (base.choices.Count>0) base.valNoCallback = base.choices[0];
        }

        public void ResetValFromChoices()
        {
            ResetValToFirstChoice();
/*
            switch (valType)
            {
                case topEnumValFlag:
                    if (!topEnumChoicesDict.Values.Contains(displayVal)) ResetValToFirstChoice();
                    break;
                case mainValFlag:
                    if (!mainChoices.Contains(mainVal)) ResetValToFirstChoice();
                    break;
                case bottomEnumValFlag:
                    if (!bottomEnumChoicesDict.Values.Contains(displayVal)) ResetValToFirstChoice();
                    break;
            }
*/
        }

        protected void ResetValToFirstChoice()
        {
            if (topEnumChoicesDict != null && topEnumChoicesDict.Keys.Count > 0) valTopEnum = topEnumChoicesDict.Keys.First();
            else if (mainChoices != null && mainChoices.Count > 0) mainVal = mainChoices[0];
            else if (bottomEnumChoicesDict != null && bottomEnumChoicesDict.Keys.Count > 0) valBottomEnum = bottomEnumChoicesDict.Keys.First();
        }

        protected void ResetChoices()
        {
            List<string> choices = new List<string>();
            List<string> displayChoices = new List<string>();
            defaultVal = "";

            foreach (KeyValuePair<int, string> kvp in topEnumChoicesDict)
            {
                if (defaultVal == "") defaultVal = topPrefix + kvp.Key.ToString();
                choices.Add(topPrefix + kvp.Key.ToString());
                displayChoices.Add(kvp.Value);
            }

            foreach (string choice in mainChoices)
            {
                if (defaultVal == "") defaultVal = mainPrefix + choice;
                choices.Add(mainPrefix + choice);
            }
            foreach (string displayChoice in mainDisplayChoices)
            {
                displayChoices.Add(displayChoicePreFix + displayChoice + displayChoicePostFix);
            }

            foreach (KeyValuePair<int, string> kvp in bottomEnumChoicesDict)
            {
                if (defaultVal == "") defaultVal = bottomPrefix + kvp.Key.ToString();
                choices.Add(bottomPrefix + kvp.Key.ToString());
                displayChoices.Add(kvp.Value);
            }

            base.choices = choices;
            base.displayChoices = displayChoices;
        }

        public int valType
        {
            get
            {
                if (base.val.StartsWith(topPrefix)) return topEnumValFlag;
                if (base.val.StartsWith(mainPrefix)) return mainValFlag;
                if (base.val.StartsWith(bottomPrefix)) return bottomEnumValFlag;
                return (unrecognisedValFlag);
            }
        }

        public string mainVal
        {
            get
            {
                if (base.val.Length < 1) return "";
                return (base.val.Substring(1, base.val.Length-1));
            }
            set
            {
                base.val = mainPrefix + value;
            }
        }

        public string mainValNoCallback
        {
            get
            {
                if (base.valNoCallback.Length < 1) return "";
                return (base.valNoCallback.Substring(1, base.valNoCallback.Length - 1));
            }
            set
            {
                base.valNoCallback = mainPrefix + value;
            }
        }


        public string displayVal
        {
            get
            {
                if (base.val.StartsWith(mainPrefix)) return displayChoicePreFix + base.val.Substring(1, base.val.Length - 1) + displayChoicePostFix;
                else
                {
                    int result;
                    if (!int.TryParse(base.val.Substring(1, base.val.Length - 1), out result)) return "";
                    if (base.val.StartsWith(bottomPrefix)) return bottomEnumChoicesDict[result];
                    else if (base.val.StartsWith(topPrefix)) return topEnumChoicesDict[result];
                }
                return "";
            }
        }

        public int valTopEnum
        {
            get
            {
                int result;
                if (!base.val.StartsWith(topPrefix) || !int.TryParse(base.val.Substring(1, base.val.Length - 1), out result)) return -1;
                return result;
            }
            set
            {
                if (!topEnumChoicesDict.ContainsKey(value)) SuperController.LogError("UIAssist.JSONStorableMultiEnumStringChooser.valTopEnum: Attempt to set enum '"+name+"' with invalid int");
                else base.val = topPrefix+value.ToString();
            }
        }

        public int valTopEnumNoCallback
        {
            get
            {
                return valTopEnum;
            }
            set
            {
                if (!topEnumChoicesDict.ContainsKey(value)) SuperController.LogError("UIAssist.JSONStorableMultiEnumStringChooser.valTopEnumNoCallback: Attempt to set enum '" + name + "' with invalid int");
                else base.valNoCallback = topPrefix + value.ToString();
            }
        }

        public int valBottomEnum
        {
            get
            {
                int result;
                if (!base.val.StartsWith(bottomPrefix) || !int.TryParse(base.val.Substring(1, base.val.Length - 1), out result)) return (-1);
                return result;
            }
            set
            {
                if (!bottomEnumChoicesDict.ContainsKey(value)) SuperController.LogError("UIAssist.JSONStorableMultiEnumStringChooser.valBottomEnum: Attempt to set enum '" + name + "' with invalid int");
                else base.val = bottomPrefix + value.ToString();
            }
        }

        public int valBottomEnumNoCallback
        {
            get
            {
                return valBottomEnum;
            }
            set
            {
                if (!bottomEnumChoicesDict.ContainsKey(value)) SuperController.LogError("UIAssist.JSONStorableMultiEnumStringChooser.valBottomEnumNoCallback: Attempt to set enum '" + name + "' with invalid int");
                else base.valNoCallback = bottomPrefix + value.ToString();
            }
        }

        public void SetMainChoices(List<string> _mainChoices, List<string> _mainDisplayChoices = null, bool resetVal = false)
        {
            if (_mainChoices == null) mainChoices = new List<string>();
            else mainChoices = _mainChoices;

            if (_mainDisplayChoices == null) mainDisplayChoices = mainChoices;
            else mainDisplayChoices = _mainDisplayChoices;
            ResetChoices();

            if (resetVal && base.val.StartsWith(mainPrefix) && mainChoices.Count>0)
            {
                if (!mainChoices.Contains(mainVal)) mainVal = mainChoices[0];
            }
        }

        public void SetTopEnum (string enumManifestName, List<int> enumChoicesExclusions = null, bool resetVal = false)
        {
            if (enumManifestName == null) enumManifestName = "";
            topEnumChoicesDict = EnumManifest.GetEnumsDisplayValDict(enumManifestName);
            topEnumStoreValsDict = EnumManifest.GetEnumsStoreValDict(enumManifestName);
            topStoreValsEnumDict = EnumManifest.GetStoreValEnumsDict(enumManifestName);

            if (enumChoicesExclusions != null)
            {
                foreach (int enumExclusion in enumChoicesExclusions)
                {
                    topEnumChoicesDict.Remove(enumExclusion);
                }
            }

            ResetChoices();

            if (resetVal && base.val.StartsWith(topPrefix) && topEnumChoicesDict.Keys.Count>1)
            {
                valTopEnum = topEnumChoicesDict.First().Key ;
            }
        }

        public void SetBottomEnum(string enumManifestName, bool resetVal = false)
        {
            if (enumManifestName == null) enumManifestName = "";
            bottomEnumChoicesDict = EnumManifest.GetEnumsDisplayValDict(enumManifestName);
            bottomEnumStoreValsDict = EnumManifest.GetEnumsStoreValDict(enumManifestName);
            bottomStoreValsEnumDict = EnumManifest.GetStoreValEnumsDict(enumManifestName);
            ResetChoices();

            if (resetVal && base.val.StartsWith(bottomPrefix) && bottomEnumChoicesDict.Keys.Count > 1)
            {
                valBottomEnum = bottomEnumChoicesDict.First().Key;
            }
        }

        public override bool StoreJSON(JSONClass jc, bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
        {
            bool flag = base.StoreJSON(jc, includePhysical, includeAppearance, forceStore);
            if (flag)
            {
                if (valType==topEnumValFlag) jc[name] = topPrefix+topEnumStoreValsDict[int.Parse(mainVal)];
                else if (valType == bottomEnumValFlag) jc[name] = bottomPrefix+bottomEnumStoreValsDict[int.Parse(mainVal)];
                else if (valType == mainValFlag) jc[name] = base.val;
            }

            return flag;
        }

        public override void RestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
        {

            //            base.RestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);           

            if (!NeedsRestore(jc, restorePhysical, restoreAppearance)) return;

            if (jc[name] != null && jc[name].Value!="")
            {
                string value = jc[name];
                if (value.StartsWith(topPrefix)) valTopEnum = topStoreValsEnumDict[value.Substring(1, value.Length - 1)];
                else if (value.StartsWith(bottomPrefix)) valBottomEnum = bottomStoreValsEnumDict[value.Substring(1, value.Length - 1)];
                else if (value.StartsWith(mainPrefix)) mainVal = value.Substring(1, value.Length - 1);
                else if (UIAConsts.debugLogging) SuperController.LogMessage("UIA.JSONStorableMultiEnumStringChooser.RestoreFromJSON: Unrecognised enumeration format for " + name + " of '"+value+"'");
            }
            else
            {
                if (!setMissingToDefault) return;
                base.val = defaultVal;
            }
        }

        public void SetDisplayChoicePreFix (string prefix)
        {
            displayChoicePreFix = prefix;
            ResetChoices();
        }
        public void SetDisplayChoicePostFix(string postfix)
        {
            displayChoicePostFix = postfix;
            ResetChoices();
        }

    }

    public class JSONStorableEnumStringChooser : JSONStorableStringChooser
    {
        public delegate void SetEnumCallback(int enumValue);
        public delegate void SetDisplayValCallback(string displayValue);

        public SetEnumCallback setEnumCallbackFunction;
        public SetDisplayValCallback setDisplayValCallbackFunction;

        protected SortedDictionary<int, string> enumChoicesDict;
        protected Dictionary<int, string> enumStoreValsDict;
        protected Dictionary<string, int> storeValsEnumDict;

        public JSONStorableEnumStringChooser(string paramName, string enumManifestName, int enumStartingVal, string displayName, List<int> enumChoicesExclusions=null) : base(paramName, null, null, displayName)
        {
            
            base.defaultVal = enumStartingVal.ToString();
            base.val = base.defaultVal;

            SetEnumChoices(enumManifestName, enumChoicesExclusions);
        }

        public JSONStorableEnumStringChooser(string paramName, string enumManifestName, int enumStartingVal, string displayName, SetEnumCallback callback, List<int> enumChoicesExclusions = null) : base(paramName, null, null, displayName)
        {
            
            base.defaultVal = enumStartingVal.ToString();
            base.val = base.defaultVal;
            SetEnumChoices(enumManifestName, enumChoicesExclusions);
            setEnumCallbackFunction = callback;
            setCallbackFunction = StringEnumCallback;
        }

        public JSONStorableEnumStringChooser(string paramName, string enumManifestName, int enumStartingVal, string displayName, SetDisplayValCallback callback, List<int> enumChoicesExclusions = null) : base(paramName, null, null, displayName)
        {
            
            base.defaultVal = enumStartingVal.ToString();
            base.val = base.defaultVal;
            SetEnumChoices(enumManifestName, enumChoicesExclusions);
            setDisplayValCallbackFunction = callback;
            setCallbackFunction = StringEnumCallback;
        }

        protected void StringEnumCallback (string value)
        {
            int result ;
            if (!int.TryParse(value, out result)) return;
            setEnumCallbackFunction?.Invoke(result);
            setDisplayValCallbackFunction?.Invoke(displayVal);
        }
        

        public void SetEnumChoices(string enumManifestName, List<int> enumChoicesExclusions = null, bool forceValToChoice = true, bool suppressCallback=false)
        {
            List<string> valList = new List<string>();
            List<string> displayValList = new List<string>();
            enumChoicesDict = EnumManifest.GetEnumsDisplayValDict(enumManifestName);
            enumStoreValsDict = EnumManifest.GetEnumsStoreValDict(enumManifestName);
            storeValsEnumDict = EnumManifest.GetStoreValEnumsDict(enumManifestName);

            foreach (KeyValuePair<int, string> kvp in enumChoicesDict)
            {
                if (enumChoicesExclusions==null || !enumChoicesExclusions.Contains(kvp.Key))
                {
                    valList.Add(kvp.Key.ToString());
                    displayValList.Add(kvp.Value);
                }
            }
            base.choices = valList;
            base.displayChoices = displayValList;

            if (forceValToChoice && !base.choices.Contains(base.val))
            {
                if (!suppressCallback) base.val = base.choices[0];
                else base.valNoCallback = base.choices[0];
                
            }
            
        }

        public new int val {
            get
            {
                int result;
                if (!int.TryParse(base.val, out result)) return (-1);
                return result;
            }
            set
            {
                if (!enumChoicesDict.ContainsKey(value)) SuperController.LogError("UIAssist.JSONStorableEnumStringChooser.val: Attempt to set enum '" + name + "' with invalid int: "+value.ToString());
                else base.val = value.ToString();
            }
        }

        public new int valNoCallback
        {
            get
            {
                return val;
            }
            set
            {
                if (!enumChoicesDict.ContainsKey(value)) SuperController.LogError("UIAssist.JSONStorableEnumStringChooser.valNoCallback: Attempt to set enum '" + name + "' with invalid int: " + value.ToString());
                else base.valNoCallback = value.ToString();
            }
        }

        public string displayVal
        {
            get
            {
                if (!enumChoicesDict.ContainsKey(val)) return ("");
                return (enumChoicesDict[val]);
            }
        }

        public new List<string> choices
        {
            get
            {
                return base.choices;
            }
        }

        public new List<string> displayChoices
        {
            get
            {
                return base.displayChoices;
            }
        }

        public new int defaultVal
        {
            get
            {
                int result;
                if (!int.TryParse(base.defaultVal, out result)) return (-1);
                return result;
            }
            set
            {
                if (!enumChoicesDict.ContainsKey(value)) SuperController.LogError("UIAssist.JSONStorableEnumStringChooser.defaultVal: Attempt to set enum '" + name + "' with invalid int: " + value.ToString());
                else base.defaultVal = value.ToString();
            }
        }

        public override bool StoreJSON(JSONClass jc, bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
        {
            bool flag = base.StoreJSON(jc, includePhysical, includeAppearance, forceStore);
            if (flag) jc[name] = enumStoreValsDict[int.Parse(base.val)];

            return flag;
        }

        public override void RestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
        {
 //           base.RestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
            if (!NeedsRestore(jc, restorePhysical, restoreAppearance)) return;
             
            if (jc[name] != null && storeValsEnumDict.ContainsKey(jc[name])) base.val = storeValsEnumDict[jc[name]].ToString();
            else
            {
                if (UIAConsts.debugLogging && jc[name]!=null && jc[name].Value!="") SuperController.LogMessage("UIA.JSONStorableEnumStringChooser.RestoreFromJSON: Unrecognised enumeration '"+ jc[name].Value +"' for " + name);
                if (!setMissingToDefault) return;
                val = defaultVal;
            }

        }
    }

    public class JSONStorableObject
    {
        protected OrderedDictionary jsParams;
        protected Dictionary<string, bool> jsParamsCopyable;

        public JSONStorableObject()
        {
            jsParams = new OrderedDictionary();
            jsParamsCopyable = new Dictionary<string, bool>();
        }
        public void ClearAllRegisteredParams()
        {
            jsParams.Clear();
            jsParamsCopyable.Clear();
        }

        public void RegisterParam(JSONStorableParam jsParam, bool storable =true, bool copyable =true)
        {
            if (jsParams.Contains(jsParam.name)) SuperController.LogError("UIAssist: Attempt to register JSONStorableParam '"+ jsParam.name+"' with storeID that already exists.");
            else
            {
                jsParams.Add(jsParam.name, jsParam);
                jsParamsCopyable.Add(jsParam.name, copyable);
                jsParam.isStorable = storable;
                jsParam.isRestorable = storable;
            }

            if (jsParam.type == JSONStorable.Type.String) (jsParam as JSONStorableString).setJSONCallbackFunction = jsonStringCallback;
            if (jsParam.type == JSONStorable.Type.StringChooser) (jsParam as JSONStorableStringChooser).setJSONCallbackFunction = jsonStringChooserCallback;
            if (jsParam.type == JSONStorable.Type.Bool) (jsParam as JSONStorableBool).setJSONCallbackFunction = jsonBoolCallback;
            if (jsParam.type == JSONStorable.Type.Float) (jsParam as JSONStorableFloat).setJSONCallbackFunction = jsonFloatCallback;
            if (jsParam.type == JSONStorable.Type.Color) (jsParam as JSONStorableColor).setJSONCallbackFunction = jsonColorCallback;
            if (jsParam.type == JSONStorable.Type.Url) (jsParam as JSONStorableUrl).setJSONCallbackFunction = jsonStringCallback;
            if (jsParam.type == JSONStorable.Type.Vector3) (jsParam as JSONStorableVector3).setJSONCallbackFunction = jsonVector3Callback;
        }

        public virtual JSONStorableParam GetJSONParam(string paramName)
        {
            if (!jsParams.Contains(paramName)) return null;
            return (JSONStorableParam) jsParams[paramName];
        }

        public ICollection GetParamList()
        {
            return jsParams.Values ;
        }

        public virtual JSONClass GetJSON(HashSet<JSONStorableParam> jspLoadExclusions=null)
        {
            JSONClass jc = new JSONClass();
            foreach (DictionaryEntry jspDictEntry in jsParams)
            {                
                JSONStorableParam jsp = (JSONStorableParam)jspDictEntry.Value;
                if (jspLoadExclusions == null || !jspLoadExclusions.Contains(jsp)) jsp.StoreJSON(jc, true, true, true);
            }

            return (jc);
        }

        public virtual void RestoreFromJSON(JSONClass jc, string _paramName = null)
        {
            foreach (DictionaryEntry jspDictEntry in jsParams)
            {
                if (_paramName==null || _paramName == jspDictEntry.Key.ToString())
                {                   
                    JSONStorableParam jsp = (JSONStorableParam)jspDictEntry.Value;
                    jsp.RestoreFromJSON(jc);
                }
            }
        }

        public virtual void CopyFrom(JSONStorableObject jsObject)
        {
            foreach (DictionaryEntry jspDictEntry in jsParams)
            {
                JSONStorableParam jsParamCopyFrom = jsObject.GetJSONParam(jspDictEntry.Key.ToString());
                JSONClass jc = new JSONClass();
                jsParamCopyFrom.StoreJSON(jc, true, true, true);
                JSONStorableParam jsParamCopyTo = jspDictEntry.Value as JSONStorableParam;
                jsParamCopyTo.RestoreFromJSON(jc, true, true, true);
            }
        }

        private void jsonStringCallback(JSONStorableString jss)
        {
            jsonStorableParamCallback(jss);
        }

        private void jsonStringChooserCallback(JSONStorableStringChooser jssc)
        {
            jsonStorableParamCallback(jssc);
        }
        private void jsonFloatCallback(JSONStorableFloat jsf)
        {
            jsonStorableParamCallback(jsf);
        }
        private void jsonBoolCallback(JSONStorableBool jsb)
        {
            jsonStorableParamCallback(jsb);
        }
        private void jsonColorCallback(JSONStorableColor jsc)
        {
            jsonStorableParamCallback(jsc);
        }
        private void jsonVector3Callback(JSONStorableVector3 jsv)
        {
            jsonStorableParamCallback(jsv);
        }
        protected virtual void jsonStorableParamCallback(JSONStorableParam jsp)
        {

        }
    }
}
