using System;
using System.Collections.Generic;
using SimpleJSON;
using MVR.FileManagementSecure;

namespace JayJayWon
{
    public class UIAPProfile
    {
        private static string GetDefaultUIAPLoadPath()
        {
            string defaultPrimaryUIAP = UIAConsts._PluginDataSubfolderName + "\\"+ UIAConsts.DefaultPrimaryUIAPFileName;
            string defaultSecondaryUIAP = UIAConsts._PluginDataSubfolderName + "\\" + UIAConsts.DefaultSecondaryUIAPFileName;
            string defaultLegacyUIAP = UIAConsts._PluginDataSubfolderName + "\\default.uiap";

            if (UIAGlobals.isPrimaryPlugin && FileManagerSecure.FileExists(defaultPrimaryUIAP)) return defaultPrimaryUIAP;
            else if (!UIAGlobals.isPrimaryPlugin && FileManagerSecure.FileExists(defaultSecondaryUIAP)) return defaultSecondaryUIAP;
            else if (FileManagerSecure.FileExists(defaultLegacyUIAP)) return defaultLegacyUIAP;
            return ("");
        }

        public static void LoadDefaultUIAP()
        {
            string defaultPath = GetDefaultUIAPLoadPath();
            if (defaultPath != "") LoadUIAPSelected(defaultPath);
            else UIAStorables.CreateDefaultButtonGrids();
        }
        public static void LoadUIAPSelected(string file, bool closeGrids = true)
        {
            try
            {
                if (file != "" && FileManagerSecure.FileExists(file))
                {
                    GameControlUI.gridsDisplay.DestroyUIButtons();
                    JSONClass jc = (JSONClass)SuperController.singleton.LoadJSON(file);
                    UIAStorables.buttonGridsList.Clear();
                    
                    if (closeGrids) GridsControlDisplay.CloseGrids();

                    string packageName = "";
                    if (file.Contains(":") && file.IndexOf(':') > 1) packageName = file.Substring(0, file.IndexOf(':'));
                    UIAStorables.LoadJSON(jc, packageName);
                   
                    GameControlUI._currentGrid = UIAStorables.buttonGridsList[0];
                    GameControlUI.gameControlDisplayMode = GameControlDisplayModes.standard ;

                    ButtonSetupPluginUI.ResetCurrentButtonSelections();
                    GameControlUI.RefreshWristUIButtonGrid();
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public static void SaveUIAPSelected(string file)
        {
            try
            {
                if (file != "")
                {
                    if (!file.EndsWith(".uiap")) { file = file + ".uiap"; }
                    SuperController.singleton.SaveJSON(UIAStorables.GetJSON(), file);
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public static void SaveDefaultUIAP()
        {
            if (!UIAGlobals.isPrimaryPlugin) SaveUIAPSelected(UIAConsts._PluginDataSubfolderName + "\\" + UIAConsts.DefaultSecondaryUIAPFileName);
            else SaveUIAPSelected(UIAConsts._PluginDataSubfolderName+"\\" + UIAConsts.DefaultPrimaryUIAPFileName);            
        }
    }
    public class ImportUIAPGridTool
    {
        public ImportUIAPGridTool()
        {
            _uiapFileNameJSON = new JSONStorableString("uiapFileName", "", UIAPFileNameUpdated);

            _uiapGridOptionsJSON = new JSONStorableStringChooser("uiapScreenOptions", null, "", "Select Grid to Import");
        }

        private static void UIAPFileNameUpdated(string filename)
        {
            try
            {
                JSONClass uiapFile = (JSONClass)SuperController.singleton.LoadJSON(filename);
                _buttonGridsArray = uiapFile["ButtonGrids"].AsArray;

                if (uiapFile["uiapFormatVersion"] != null) _uiapFormatVersion = uiapFile["uiapFormatVersion"].AsInt;
                else _uiapFormatVersion = 1;

                List<string> choices = new List<string>();
                int gridID = 1;

                foreach (JSONClass buttonGrid in _buttonGridsArray)
                {
                    string gridLabel = buttonGrid["gridLabel"];
                    string choice = gridID.ToString();
                    if (gridLabel != "") choice = choice + " (" + gridLabel + ")";
                    choices.Add(choice);
                    gridID++;
                }

                _qlbButtonGrid = uiapFile["QuickLaunchButtonGrid"].AsObject;
                if (_qlbButtonGrid != null && PatreonFeatures.patreonContentEnabled)
                {
                    choices.Insert(0, "Quick Launch Buttons");
                }

                _uiapGridOptionsJSON.choices = choices;
                _uiapGridOptionsJSON.valNoCallback = _uiapGridOptionsJSON.choices[0];
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }

        public static void ImportGrid()
        {
            try
            {
                string packageName = "";
                if (_uiapFileNameJSON.val.Contains(":") && _uiapFileNameJSON.val.IndexOf(':') > 1) packageName = _uiapFileNameJSON.val.Substring(0, _uiapFileNameJSON.val.IndexOf(':'));

                UIAButtonGrid newBG = new UIAButtonGrid(3, 3);
                JSONClass buttonGridJSON = null;

                if (_uiapGridOptionsJSON.val == "Quick Launch Buttons" && _qlbButtonGrid != null) buttonGridJSON = _qlbButtonGrid;
                else if (_uiapGridOptionsJSON.val != "Quick Launch Buttons" && _buttonGridsArray != null)
                {
                    int screenIndex = 0;
                    if (_uiapGridOptionsJSON.val.Contains(" ("))
                    {
                        string screenIndexString = _uiapGridOptionsJSON.val.Substring(0, _uiapGridOptionsJSON.val.IndexOf(" ("));
                        int.TryParse(screenIndexString, out screenIndex);
                    }
                    else int.TryParse(_uiapGridOptionsJSON.val, out screenIndex);
                    buttonGridJSON = (JSONClass)_buttonGridsArray[screenIndex - 1];
                }

                if (buttonGridJSON != null)
                {
                    newBG.LoadJSON(buttonGridJSON, packageName, _uiapFormatVersion);

                    UIAStorables.buttonGridsList.Add(newBG);
                    int gridRef = UIAStorables.buttonGridsList.Count;
                    ButtonSetupPluginUI.gridSelectorJSSC.val = gridRef.ToString();
                    GameControlUI.RefreshWristUIButtonGrid();
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public static void ImportQLBGrid()
        {
            try
            {
                if (_uiapGridOptionsJSON.val == "Quick Launch Buttons" && _qlbButtonGrid != null)
                {
                    string packageName = "";
                    if (_uiapFileNameJSON.val.Contains(":") && _uiapFileNameJSON.val.IndexOf(':') > 1) packageName = _uiapFileNameJSON.val.Substring(0, _uiapFileNameJSON.val.IndexOf(':'));

                    UIAStorables.quickLaunchButtonGrid.LoadJSON(_qlbButtonGrid, packageName, _uiapFormatVersion);
                    ButtonSetupPluginUI.gridSelectorJSSC.val = "QL";
                    GameControlUI.RefreshWristUIButtonGrid();
                }

            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }

        public static JSONStorableString _uiapFileNameJSON;
        public static JSONStorableStringChooser _uiapGridOptionsJSON;
        private static JSONArray _buttonGridsArray = null;
        private static JSONClass _qlbButtonGrid = null;
        private static int _uiapFormatVersion;

    }


}
